from main import REGISTRY_PATH

import os
import random
import shelve
import shutil
import threading
import string


class StateSaverServer(object):
    """
      This class provides a singleton server object for saving program state
      in a database. The path of the database file contains random elements in
      order to not be attackable by "known-location attacks".
    """

    __STATES_FILE = os.path.join(REGISTRY_PATH,
              "states" + hex(random.randrange(0xffffffL, 0xffffffffL)) + ".db")
    __SEPARATOR = ' '

    def __init__(self):

        self.__lock = threading.Lock()

        self.__db_file = self.__find_db()
        self.__db = shelve.open(self.__db_file, flag = 'c', writeback = True)

        # flush the cache every 5 seconds
        import gobject
        gobject.timeout_add(5 * 1000, self.__flush)



    def __flush(self):

        self.__db.sync()
        return True



    def __find_db(self):

        files = os.listdir(REGISTRY_PATH)
        dbs = [ f for f in files if f.startswith("states") ]

        if (not dbs):
            return self.__STATES_FILE
        else:
            pos = dbs[0].find(".db")
            return os.path.join(REGISTRY_PATH, dbs[0][0:pos+3])



    def __check_key(self, key):

        assert key
        assert key[0] not in string.digits

        for c in key:
            assert c in (string.ascii_letters + string.digits + '-_.[]')



    def set_key(self, ident, key, value):

        self.__check_key(key)

        self.__lock.acquire()

        try:
            pk = ident + self.__SEPARATOR + key
            self.__db[pk] = value
        finally:
            self.__lock.release()



    def get_key(self, ident, key, default):

        self.__check_key(key)

        self.__lock.acquire()

        try:
            pk = ident + self.__SEPARATOR + key
            return self.__db.get(pk, default)
        finally:
            self.__lock.release()



    def remove(self, ident):

        self.__lock.acquire()

        try:
            for k in self.__db.keys():
                if k.startswith(ident + self.__SEPARATOR):
                    del self.__db[k]
        finally:
            self.__lock.release()



    def list(self, ident):

        self.__db.sync()
        return [k[len(ident) + 1:] for k in self.__db.keys() \
                if k.startswith(ident + self.__SEPARATOR)]


_SERVER = StateSaverServer()



class StateSaverClient(object):

    __slots__ = "__ident",

    def __init__(self, ident):
        self.__ident = ident

    def set_key(self, key, value):
        _SERVER.set_key(self.__ident, key, value)

    def get_key(self, key, default = None):
        return _SERVER.get_key(self.__ident, key, default)

    def remove(self):
        _SERVER.remove(self.__ident)

    def list(self):
        return _SERVER.list(self.__ident)


StateSaver = StateSaverClient

_singleton = StateSaver("__default__")
def DefaultStateSaver(): return _singleton



from ContainerTarget import ContainerTarget
from DisplayTarget import DisplayTarget
from utils.datatypes import *
from layout import Unit

import gtk


#
# Class for aligning targets.
#
class TargetAlignment(ContainerTarget):

    def __init__(self, name, parent):

        self.__alignment = (0.0, 0.0)

        ContainerTarget.__init__(self, name, parent)
        self.__align = gtk.Alignment(0.5, 1, 0, 0)
        self.__align.set_direction(gtk.TEXT_DIR_LTR)
        self.__align.show()

        self._register_property("align-x", TYPE_FLOAT,
                                self._setp_alignment, self._getp)
        self._register_property("align-y", TYPE_FLOAT,
                                self._setp_alignment, self._getp)


    def get_widget(self): return self.__align


    def new_child(self, childtype, settings, children):

        child = ContainerTarget.new_child(self, childtype, settings, children)
        self.__align.add(child.get_widget())



    #
    # Alignment properties.
    #
    def _setp_alignment(self, key, value):

        if (key == "align-x"):
            ax, ay = self.__alignment
            self.__align.set(value, ay, 0, 0)
            self.__alignment = (value, ay)

        elif (key == "align-y"):
            ax, ay = self.__alignment
            self.__align.set(ax, value, 0, 0)
            self.__alignment = (ax, value)

        child = self._get_child()
        nil, nil, aw, ah = self.get_geometry()
        nil, nil, cw, ch = child.get_geometry()
        hvalue, vvalue = self.__alignment
        x = (aw - cw) * hvalue
        y = (ah - ch) * vvalue
        child.set_prop("x", x)
        child.set_prop("y", y)
        
        self._setp(key, value)



from libdesklets.controls import Interface, Permission

class IURI(Interface):
    
    file     = Permission.READWRITE
    raw      = Permission.READ
    stripped = Permission.READ
    splitted = Permission.READ



from sensor.Sensor import Sensor

import commands
import time


#
# Sensor for invoking external programs.
#
class External(Sensor):

    def __init__(self, cmd, interval = 1000):

        self.__output = ""


        Sensor.__init__(self)
        self._add_timer(100, self.__on_tick, int(interval))
        if (int(interval) > 0):
            self._add_thread(self.__command_thread, cmd,
                             0.001 * float(interval))
        else:
            fail, output = commands.getstatusoutput(cmd)
            if (fail): output = ""
            self.__output = output



    def __on_tick(self, interval):

        data = self._new_output()
        if (self.__output): data.set("value", self.__output)
        self._send_output(data)

        if (interval > 0): self._add_timer(interval, self.__on_tick, interval)



    def __command_thread(self, cmd, interval):

        while not self._is_stopped():

            fail, output = commands.getstatusoutput(cmd)
            if (fail): output = ""

            self.__output = output
            time.sleep(interval)


def new_sensor(args): return External(*args)



from shell.Plugin import Plugin

import gtk


#
# View for managin controls.
#
class View_Controls(Plugin):

    def activate(self):

        self._get_plugin("UI_ControlBrowser").show()


    def deactivate(self):

        self._get_plugin("UI_ControlBrowser").hide()


    def get_name(self):

        return "Controls"


    def get_icon(self):

        img = gtk.Image()
        img.set_from_stock(gtk.STOCK_EXECUTE, gtk.ICON_SIZE_SMALL_TOOLBAR)
        img.show()
        
        return img


    def get_priority(self): return 10
    

def get_class(): return View_Controls



from plugin.ControlRegistry import ControlRegistry
from plugin.Interface import Interface
from main import CONTROLPATHS, HOME

import os
import sys


if os.curdir not in sys.path: sys.path.append(os.curdir)
if HOME not in sys.path: sys.path.append(HOME)


#
# Factory for plugins. We don't look up plugins by their implementation but
# by their interface. Several plugins providing the same interface may
# co-exist.
#
class _ControlFactory:

    def __init__(self):

        self.__after_rebuilding = False

        self.__registry = ControlRegistry(CONTROLPATHS)
        


    #
    # Returns a control providing the given interface.
    #
    def get_control(self, iface):

        ctrl = self.__registry.get_control(iface)
        if (ctrl and os.path.isdir(ctrl)):
            cwd = os.getcwd()
            os.chdir(os.path.dirname(ctrl))
            m = __import__(os.path.basename(ctrl))
            obj = m.get_class()()
            os.chdir(cwd)
            return obj
        else:
            return None



_singleton = _ControlFactory()
def ControlFactory(): return _singleton



from ConfigWidget import ConfigWidget
from utils.datatypes import *

import gtk


class ConfigBoolean(ConfigWidget):

    def __init__(self, name, getter, setter, caller):

        self.__initialized = False

        ConfigWidget.__init__(self, name, getter, setter, caller)

        self._register_property("value", TYPE_BOOL, self._setp_value,
                                self._getp, False, doc = "Boolean value")


    def get_widgets(self):

        self.__check = gtk.CheckButton("")
        self.__check.show()
        self.__check.connect("toggled", self.__on_change)

        return (self.__check,)


    def __on_change(self, src):

        value = src.get_active()
        self._set_config(value)


    def _set_label(self, value): self.__check.set_label(value)
    def _set_enabled(self, value): self.__check.set_sensitive(value)


    def _setp_value(self, key, value):

        old_value = self._getp("value")
        self.__check.set_active(value)

        if (not self.__initialized and value == old_value):
            # force "toggled" signal
            self.__check.toggled()
        self.__initialized = True
            
        self._setp(key, value)




from Generic import Generic

import re
import struct

# Also works for x86_64
# it seems that there's no difference

class X86(Generic):

    def __init__(self):

        Generic.__init__(self)

        def _get_model():
            r = re.compile('^model name\s+:\s+(.+)$', re.M)
            m = r.search( self._read_cpuinfo() )
            return m.group(1)

        def _get_cache():
            r = re.compile('^cache size\s+:\s+(\d+)\sKB$', re.M)
            m = r.search( self._read_cpuinfo() )
            return int(m.group(1))

        # CPU model and cache size never changes
        self.__model_name = _get_model()
        self.__cache_size = _get_cache()

        # the cpu speed might change (laptops have mobile CPUs)
        self.__speed      = re.compile('^cpu MHz\s+:\s+(\d+\.\d+)$', re.M)



    def cpu_cache(self):
        """
        @return : 2nd level cache of installed processor
        @rtype  : int
        """

        return self.__cache_size



    def cpu_model(self):
        """
        @return : model/type of installed processor
        @rtype  : str
        """

        return self.__model_name



    def cpu_speed(self):
        """
        @return : current clock of installed processor
        @rtype  : float
        """

        m = self.__speed.search( self._read_cpuinfo() )
        return float(m.group(1))



    def users(self):

        # man utmp
        # don't know if
        # sizeof(struct utmp) == 384
        # sizeof(short) == 2
        # on every Linux arch
        # http://gnomesupport.org/forums/viewtopic.php?p=33686
        # X86, X86_64 : ok

        count = 0
        data = open('/var/run/utmp', 'rb').read()

        for i in range(0, len(data), 384):
            ut_type = struct.unpack('h', data[i:i+2])[0]
            if (ut_type == 7): count += 1

        return count




# TODO : add epydoc

from main import _

import struct
import socket

_TERA = 2**40
_GIGA = 2**30
_MEGA = 2**20
_KILO = 2**10


def human_readable_bytes(bytes):

    # TODO: replace with gnome_vfs_file_size_format_for_display when available
    
    global _TERA, _GIGA, _MEGA, _KILO

    if bytes >= _TERA:
        return _("%.2f TB") % (float(bytes) / _TERA)
    elif bytes >= _GIGA:
        return _("%.2f GB") % (float(bytes) / _GIGA)
    elif bytes >= _MEGA:
        return _("%.2f MB") % (float(bytes) / _MEGA)
    elif bytes >= _KILO:
        return _("%.2f kB") % (float(bytes) / _KILO)
    else:
        return _("%d B") % bytes



def bytes_to_gmkb(bytes):
    global _TERA, _GIGA, _MEGA, _KILO

    gigs, bytes = divmod(bytes, _GIGA)
    megs, bytes = divmod(bytes, _MEGA)
    ks  , bytes = divmod(bytes, _KILO)

    return (gigs, megs, ks, bytes)



def gmkb_to_bytes(gigs, megs, ks, bytes):
    global _TERA, _GIGA, _MEGA, _KILO

    b  = gigs * _GIGA
    b += megs * _MEGA
    b += ks   * _KILO
    b += bytes

    return b


def secs_to_dhms(secs):

    day,  secs = divmod(secs, 86400)
    hou,  secs = divmod(secs, 3600)
    mint, secs = divmod(secs, 60)

    return (day, hou, mint, secs)



def dhms_to_secs(d, h, m, s):

    secs  = 86400 * d
    secs += 3600 * h
    secs += 60 * m
    secs += s

    return secs



def centigrade_to_fahrenheit(c):

    return (c * 1.8) + 32.0



def fahrenheit_to_centigrade(f):

    return (f - 32.0) / 1.8



def centigrade_to_kelvin(c):

    return c + 273.16



def kelvin_to_centigrade(k):

    return k - 273.16



# TODO: check for correctness


def ipv4_to_dotted_quad(ip):

    """Converts ipv4 (32bits unsigned integer) to
    dotted-quad decimal representation

    e.g:
      ipv4_to_dotted_quad(1291888832)
        -> '192.168.0.77'
    """
    return socket.inet_ntoa(struct.pack('=L', ip))



def dotted_quad_to_ipv4(ip):

    """Converts dotted-quad decimal to
     ipv4 (32bits unsigned integer) representation

    e.g:
      dotted_quad_to_ipv4('192.168.0.77')
        -> 1291888832
    """
    return struct.unpack('=L', socket.inet_aton(ip))[0]



#
# Registry for DataTargets. To add a new target just add an appropriate entry
# to _REGISTRY.
#

from TargetAlignment     import TargetAlignment
from TargetArray         import TargetArray
from TargetDisplay       import TargetDisplay
from TargetEntry         import TargetEntry
from TargetExpander      import TargetExpander
from TargetFrame         import TargetFrame
from TargetGauge         import TargetGauge
from TargetGroup         import TargetGroup
from TargetImage         import TargetImage
from TargetLabel         import TargetLabel

if (not HAVE_WIN32):
    from TargetBonoboControl import TargetBonoboControl
    from TargetCanvas        import TargetCanvas
    from TargetPlotter       import TargetPlotter

_targets = {
    "alignment" : (TargetAlignment, True),
    "array"     : (TargetArray, False),
    "display"   : (TargetDisplay, False),
    "entry"     : (TargetEntry, False),
    "expander"  : (TargetExpander, True),
    "frame"     : (TargetFrame, True),
    "gauge"     : (TargetGauge, True),
    "group"     : (TargetGroup, False),
    "image"     : (TargetImage, False),
    "label"     : (TargetLabel, False),
}

if (not HAVE_WIN32):
    _targets.update({
        "canvas"    : (TargetCanvas, False),
        "embed"     : (TargetBonoboControl, False),
        "plotter"   : (TargetPlotter, False)
        })



#
# Creates and returns the given target.
#
def create(name, parent):

    try:
        clss, one_child = _targets[name]
    except KeyError:
        raise UserError(_("Unknown element <b>&lt;%s&gt;</b>") % name,
                        _("Either there is a typo in the .display file "
                          "or you have an incompatible version of gDesklets."))

    obj = clss(name, parent)
    return obj


#
# Returns whether the given container accepts more than one children.
#
def one_child(name):

    clss, one_child = _targets[name]
    return one_child



#
# Base class for observable objects.
#
class Observable(object):

    #
    # Ensures that the observable is initialized.
    # The user doesn't have to call the constructor explicitly.
    #
    def __ensure_init(self):

        try:
            self.__handlers
        except:
            self.__handlers = []



    def add_observer(self, observer):

        self.__ensure_init()
        self.__handlers.append(observer)



    def remove_observer(self, observer):

        self.__ensure_init()
        self.__handlers.remove(observer)



    def drop_observers(self):

        self.__ensure_init()
        self.__handlers = []



    def update_observer(self, *args):

        self.__ensure_init()

        for h in self.__handlers[:]:
            h(self, *args)



# from libdesklets.system.gtop import _Struct as Struct

class Struct(dict):

    """
    Class for smart handling a bunch of data.

    The following is a example
      >>> s = Struct(spam=42, egg=range(3))
      >>> print s, s.spam, s.egg
      {'egg': [0, 1, 2], 'spam': 42} 42 [0, 1, 2]

      >>> stat = get_system_information()
      >>> display(stat.info1)

    """

    __slots__ = ()

    def __getattr__(self, name):

        return self[name]


#     def __getattribute__(self, name):

#         try:
#             return self[name]
#         except KeyError:
#             return dict.__getattribute__(self, name)

    def __repr__(self):

        s = 'Struct {'

        items = self.items()
        items.sort()

        for k, v in items:

            s += ' .%s = %s,' % (k, v)

        if s[-1] == ',':
            s = s[:-1]

        s += ' }'

        return s


    def __hash__(self):

        return hash(repr(self))
    




#
# Library for common functionality.
#

# convert
#   human_readable_bytes(bytes:int): (int)
#   bytes_to_gmkb(bytes:int): (int, int, int, int)
#   gmkb_to_bytes(GB:int, MB:int, kB:int, Bytes:int): (bytes:int)
#   secs_to_dhms(secs:int): (days:int, hrs:int, mins:int, secs:int)
#   dmhs_to_secs(days:int, hrs:int, mins:int, secs:int)
#   centigrade_to_fahrenheit(centigrade:int): (fahrenheit:int)
#   fahrenheit_to_centigrade(fahrenheit:int): (centigrade:int)
#   centigrade_to_kelvin(centigrade:int): (kelvin:int)
#   kelvin_to_centigrade(kelvin:int): (centigrade:int)

from convert import *

from utils.Struct import Struct



from config import settings

class _UNIT(str): pass

# for now, we can't use _UNIT() yet; there is some code to be modified first
UNIT_CM = "cm" #_UNIT("cm")
UNIT_IN = "in" #_UNIT("in")
UNIT_PERCENT = "%" #_UNIT("%")
UNIT_PT = "pt" #_UNIT("pt")
UNIT_PX = "px" #_UNIT("px")


def _round(v): return (v < 0) and int(v - 0.5) or int(v + 0.5)


class Unit(object):
    """
      Class for coordinate and size values with units. This class abstracts from
      actual lengths with concrete units.
    """

    __slots__ = ('__is_unset', '__unit', '__cache_px', '__cache_pt',
                 '__cache_cm', '__cache_in', '__cache_pcnt', '__in2pcnt',
                 '__original_value', '__value',
                 '__IN2PT_SCALE', '__PT2IN_SCALE', '__IN2PX_SCALE',
                 '__PX2IN_SCALE', '__IN2CM_SCALE', '__CM2IN_SCALE')


    def __init__(self, value = None, unit = None, string = None):

        # scaling factors
        self.__IN2PT_SCALE = 72.0
        self.__PT2IN_SCALE = 1 / self.__IN2PT_SCALE
        self.__IN2PX_SCALE = float(settings.dpi)
        self.__PX2IN_SCALE = 1 / self.__IN2PX_SCALE
        self.__IN2CM_SCALE = 2.54
        self.__CM2IN_SCALE = 1 / self.__IN2CM_SCALE

        if (not string): string = None
        if (value == unit == string == None):
            value, unit = 0, UNIT_PX
            self.__is_unset = True
        else:
            self.__is_unset = False

        if (string):
            if (string[-1] == "%"):
                value = float(string[:-1])
                unit = UNIT_PERCENT
            elif (string[-2:] == "cm"):
                value = float(string[:-2])
                unit = UNIT_CM
            elif (string[-2:] == "in"):
                value = float(string[:-2])
                unit = UNIT_IN
            elif (string[-2:] == "pt"):
                value = float(string[:-2])
                unit = UNIT_PT
            else:
                value = int(float(string))
                unit = UNIT_PX


        # the unit type
        self.__unit = unit

        # cache for computed values
        self.__cache_px = (unit != UNIT_PX) and value or None
        self.__cache_pt = (unit == UNIT_PT) and value or None
        self.__cache_cm = (unit == UNIT_CM) and value or None
        self.__cache_in = (unit == UNIT_IN) and value or None
        self.__cache_pcnt = (unit == UNIT_PERCENT) and value or None

        # the scaling factor from inches to percents
        self.__in2pcnt = 0.1

        self.__validate(self.__unit)

        # store value in inches
        self.__original_value = value
        self.__value = self.__unit_to_inch(value, unit)
        self.set_100_percent(100)
        


    #
    # Validates the unit type.
    #
    def __validate(self, unit):

        if (unit not in (UNIT_PX, UNIT_CM, UNIT_IN, UNIT_PT, UNIT_PERCENT)):
            raise TypeError("%s is not a valid UNIT." % (unit,))



    def __unit_to_inch(self, value, unit):

        # trust no one
        self.__validate(unit)

        if (unit == UNIT_IN): return value
        elif (unit == UNIT_CM): return value * self.__CM2IN_SCALE
        elif (unit == UNIT_PX): return value * self.__PX2IN_SCALE
        elif (unit == UNIT_PT): return value * self.__PT2IN_SCALE
        # percentual values get converted at a later time
        elif (unit == UNIT_PERCENT): return value


    def __inch_to_unit(self, value, unit):

        # trust no one
        self.__validate(unit)

        if (unit == UNIT_IN):
            v = value
            self.__cache_in = v
        elif (unit == UNIT_CM):
            v = value * self.__IN2CM_SCALE
            self.__cache_cm = v
        elif (unit == UNIT_PX):
            v = _round(value * self.__IN2PX_SCALE)
            self.__cache_px = v
        elif (unit == UNIT_PT):
            v = value * self.__IN2PT_SCALE
            self.__cache_pt = v
        elif (unit == UNIT_PERCENT):
            v = value * self.__in2pcnt
            self.__cache_pcnt = v

        return v


    def set_100_percent(self, size):
        #assert (size > 0)
        if (size <= 0): size = 1

        # size is a value in pixels, therefore convert it to inches
        size = self.__unit_to_inch(size, UNIT_PX)

        if (self.__unit == UNIT_PERCENT):
            self.__value = (self.__original_value) * (size / 100.0)

        self.__in2pcnt = 100.0 / size

        # invalidate cached values
        self.__cache_pcnt = None
        self.__cache_px = None
        self.__cache_pt = None
        self.__cache_cm = None
        self.__cache_in = None



    def as_px(self):
        return self.__cache_px or self.__inch_to_unit(self.__value, UNIT_PX)

    def as_pt(self):
        return self.__cache_pt or self.__inch_to_unit(self.__value, UNIT_PT)

    def as_cm(self):
        return self.__cache_cm or self.__inch_to_unit(self.__value, UNIT_CM)

    def as_in(self):
        return self.__cache_in or self.__inch_to_unit(self.__value, UNIT_IN)

    def as_percent(self):
        return self.__cache_pcnt or self.__inch_to_unit(self.__value,
                                                        UNIT_PERCENT)


    def get_unit(self): return self.__unit
    def is_unset(self): return self.__is_unset



    #
    # Creates a new Unit object as a copy of this one.
    # TODO: copy the set_100_percent value as well
    #
    def copy(self):

        new = Unit(self.as_pt(), UNIT_PT)
        return new


    def __getitem__(self, unit):

        log("Deprecation:\n"
            "Accessing the Unit object as a dictionary is deprecated. "
            "Please use the appropriate accessor methods instead.", True)

        return self.__inch_to_unit(self.__value, unit)


    def __add__(self, other):

        return Unit(self.as_pt() + other.as_pt(), UNIT_PT)


    def __sub__(self, other):

        return Unit(self.as_pt() - other.as_pt(), UNIT_PT)


    def __mul__(self, other):

        if (isinstance(other, Unit)):
            return Unit(self.as_pt() * other.as_pt(), UNIT_PT)
        else:
            return Unit(self.as_pt() * other, UNIT_PT)


    def __div__(self, other):

        if (isinstance(other, Unit)):
            return Unit(self.as_pt() / other.as_pt(), UNIT_PT)
        else:
            return Unit(self.as_pt() / other, UNIT_PT)


    def __cmp__(self, other):
        assert (other == None or isinstance(other, Unit)), \
               "Cannot compare with non-Unit object"

        if (not other): return 1
        else: return cmp(int(self.as_pt()), int(other.as_pt()))


    def __repr__(self):
        unit = self.get_unit()
        if (self.is_unset()): return "unset"
        if (unit == UNIT_CM): return "%fcm" % self.as_cm()
        elif (unit == UNIT_IN): return "%fin" % self.as_in()
        elif (unit == UNIT_PT): return "%fpt" % self.as_pt()
        elif (unit == UNIT_PX): return "%f" % self.as_px()
        elif (unit == UNIT_PERCENT): return "%f%%" % self.as_percent()


    def __hash__(self):

        return hash(repr(self))

import sys
# for convenience
ZERO = Unit(0, UNIT_PT)
ONE = Unit(1, UNIT_PT)
MAXIMUM = Unit(sys.maxint, UNIT_PT)
UNSET = Unit()


if (__name__ == "__main__"):

    u = Unit(10, UNIT_PX)
    u.set_100_percent(11)
    print u.as_px()
    print u.as_in()
    print u.as_percent()



from ContainerTarget import ContainerTarget
from utils.tiling import Tiling as TilingImage
from utils.datatypes import *
import utils

import gtk
import os


#
# Class for grouping different targets together.
#
class TargetGroup(ContainerTarget):

    def __init__(self, name, parent):

        # flag for ensuring that only one collapse action takes place
        self.__block = 0

        # the maximum size values of all the children
        self.__max_size = (0, 0)

        ContainerTarget.__init__(self, name, parent)
        self.__layout = gtk.Fixed()
        self.__layout.show()

        self.__image = TilingImage()
        self.__image.show()
        self.__layout.put(self.__image, 0, 0)

        self._register_property("bg-uri", TYPE_STRING,
                                self._setp_bg_uri, self._getp)
        self._register_property("bg-color", TYPE_STRING,
                                self._setp_bg_color, self._getp)

        self.add_observer(self.__on_observe_size)



    def delete(self):

        self.__layout.remove(self.__image)
        del self.__image
        del self.__layout
        del self.__block
        del self.__max_size
        ContainerTarget.delete(self)



    def get_widget(self): return self.__layout


    def __on_observe_size(self, src, cmd):

        if (cmd == src.OBS_GEOMETRY):
            x, y, w, h = self.get_geometry()

            if (self._getp("bg-uri") or self._getp("bg-color")):
                ow, oh = self.__image.size_request()
                if ((w.as_px(), h.as_px()) != (ow, oh)):
                    self.__image.tile(w.as_px(), h.as_px())



    def new_child(self, childtype, settings, children):

        child = ContainerTarget.new_child(self, childtype, settings, children)
        if (not child.is_standalone()):
            x, y, w, h = child.get_geometry()
            self.__layout.put(child.get_widget(), x.as_px(), y.as_px())

        return child



    def child_observer(self, src, cmd):

        if (cmd == src.OBS_GEOMETRY):
            x, y, w, h = src.get_geometry()

            if (src.get_widget() in self.__layout.get_children()):
                self.__layout.move(src.get_widget(), x.as_px(), y.as_px())



    #
    # Sets the background color.
    #
    def __set_color(self, color):

        r, g, b, a = utils.parse_color(color)
        w, h = self.__layout.size_request()
        self.__image.set_from_color(r, g, b, a)
        self.__image.tile(w, h)



    #
    # Sets the background image.
    #
    def __set_background(self, uri):

        if (not uri):
            self.__set_color("#00000000")
            return

        from utils import vfs
        if (vfs.exists(uri)):
            try:
                data = vfs.read_entire_file(uri)
            except:
                return
            w, h = self.__layout.size_request()
            self.__image.set_from_data(data)
            self.__image.tile(w, h)



    #
    # "bg-uri" property.
    #
    def _setp_bg_uri(self, key, value):

        path = self._get_display().get_full_path(value)
        self.__set_background(path)
        self._setp(key, value)



    #
    # "bg-color" property.
    #
    def _setp_bg_color(self, key, value):

        self.__set_color(value)
        self._setp(key, value)




from plugin.Interface import Interface
import utils

import gobject

class Control(object):

    """
    Base class for controls. Every control has to be derived from this class.
    """

    # Warning : this is the only way to escape a Control
    # so be carefull when allowing a method to be called

    AUTHORIZED_METHODS = ('bind', )

    def __init__(self):

        # the handlers for property bindings
        self.__handlers = {}

        # flag indicating whether the control has been stopped
        self.__is_stopped = False

        # the path of the display file (excluding the file)
        self.__display_path = ""
        

    def _update(self, prop):

        """ Issues a change of the given property. """

        lst = self.__handlers.get(prop, [])
        for handler, args in lst:
            utils.run_in_main_thread(handler, getattr(self, prop), *args)


    def bind(self, prop, handler, *args):

        """ Binds a handler to property changes. """

        self.__handlers.setdefault(prop, []).append((handler, args))


    def _add_timer(self, interval, callback, *args):

        """ Runs a timer callback at the given intervals. """

        def func():
            """ callback function for gobject.timeout_add """
            if self.__is_stopped:
                return False
            return bool(callback(*args))

        return gobject.timeout_add(interval, func)


    def _remove_timer(self, ident):

        """ Removes timer with given ID. """

        gobject.source_remove(ident)


    def stop(self):

        """ Stops the control. """

        self.__is_stopped = True
        # remove bound handlers
        self.__handlers.clear()
        self._shutdown()


    def _shutdown(self):

        """
        Shutdown handler for controls. Controls should override this
        method if needed.
        """

        pass


    def __interface(self):

        """
        Returns a description of the interface of this control.
        """

        return Interface.text_describe(self.__class__)

    interface = property(__interface, doc = "Interface description")




from utils.datatypes import *


class Element(object):
    """
      Base class for elements with typed properties.
      Classes deriving from this base class have to call

        _register_property(name, datatype, setter, getter, default, doc)

      in the constructor. Properties can then be set with

        set_prop(key, value)

      and can be read with

        get_prop(key, value)


      If you don't need any special setter or getter methods, then you can use
      the predefined _setp(key, value) and _getp(key) methods.
    """

    __slots__ = ('__name', '__properties', '__property_handlers',
                 '__ident_counter')

    AUTHORIZED_METHODS = ()

    # counter for unique IDs
    __ident_counter = 0


    def __init__(self, name):

        # name of the element
        self.__name = name

        # values of the properties
        self.__properties = {}

        # table: property name -> (setter, getter, datatype, doc)
        self.__property_handlers = {}


        new_id = "id%d" % self.__ident_counter
        self.__ident_counter += 1
        self._register_property("id", TYPE_STRING, self._setp, self._getp,
                                new_id, doc = "Unique identifier")



    #
    # Registers the given property with the given setter and getter methods.
    #
    def _register_property(self, name, datatype, setter, getter,
                           default = None, doc = ""):

        self.__property_handlers[name] = (setter, getter, datatype, doc)
        self._setp(name, default)



    #
    # Sets the given property as a string.
    #
    def set_prop_from_string(self, key, value):

        value = str(value)  # to stay compatible with sensors :(
        key = key.replace("_", "-")
        try:
            datatype = self.get_datatype_of_property(key)
        except KeyError:
            datatype = TYPE_ANY

        from utils import typeconverter
        self.set_prop(key, typeconverter.str2type(datatype, value))



    #
    # Sets the given property.
    #
    def set_prop(self, key, value):

        key = key.replace("_", "-")
        try:
            setter = self.__property_handlers[key][0]
            datatype = self.__property_handlers[key][2]

        except KeyError:
            raise UserError(_("No such property: %s") % key,
                            _("The element <b>%s</b> does not have the "
                              "<b>%s</b> property.") % (self.__name, key))
        
        if (not setter):
            raise UserError(_("Permission Error"),
                           _("The property <b>%s</b> of element <b>%s</b> "
                             "is not writable.") % (key, self.__name))
        
        elif (dtype_check(datatype, value)):
            setter(key, value)

        else:
            actual_type = dtype_guess(value)
            raise UserError(_("Type Error"),
                           _("The property <b>%s</b> of element <b>%s</b> "
                             "got a value of wrong type.\n"
                             "Expected <b>%s</b>, but got <b>%s</b>."
                             % (key, self.__name, datatype[0],
                                actual_type[0])))



    #
    # Returns the value of the given property.
    #
    def get_prop(self, key):

        key = key.replace("_", "-")
        try:
            getter = self.__property_handlers[key] [1]
        except KeyError:
            raise KeyError("Error: No such property: %s" % key)

        if (not getter):
            raise UserError(_("Permission Error"),
                           _("The property <b>%s</b> of element <b>%s</b> "
                             "is not readable.") % (key, self.__name))
        
        else:
            return getter(key)



    #
    # Returns the datatype of the given property.
    #
    def get_datatype_of_property(self, prop):

        # we intentionally don't catch the KeyError here
        return self.__property_handlers[prop][2]



    #
    # Returns the documentation string of the given property.
    #
    def get_doc_of_property(self, prop):

        try:
            doc = self.__property_handlers[prop][3]
        except:
            return ""

        return doc



    #
    # Returns a list of the names of all available properties.
    #
    def get_props(self): return self.__property_handlers.keys()



    #
    # Generic setter and getter methods for properties.
    #
    def _setp(self, key, value): self.__properties[key] = value
    def _getp(self, key): return self.__properties[key]


    #
    # Returns the element's name
    #
    def get_name(self): return self.__name



from StateSaver import StateSaver
from utils.Observable import Observable
import utils


class _ConfigManager(Observable):
    """
    This class acts as a compatibility adaptor for deprecated sensors and can
    be removed as soon as support for sensors is being dropped.
    """

    UNDEF = "-- undef --"

    OBS_UPDATE = 1
    

    def __init__(self):

        self.__db = StateSaver("sensors")

        # hashtable for config change handlers: path -> handler
        self.__handlers = {}

        self.add_observer(self.__on_observe_backend)



    def __on_observe_backend(self, src, cmd, *args):

        if (cmd == src.OBS_UPDATE):
            path, value = args

            parts = path.split(".")
            for i in range(len(parts)):
                handler = self.__handlers.get(tuple(parts[:i + 1]))
            
                if (handler):
                    handler(parts, value)
                    break
            #end for



    def watch(self, *args):

        """Sets a callback handler for watching changes for the given
        configuration entry."""

        path = args[:-1]
        print "ADD WATCH", path
        handler = args[-1]
        self.__handlers[path] = handler



    def remove_watcher(self, *args):

        """Removes a watch callback handler."""

        try:
            path = args[:-1]
            del self.__handlers[path]
        except KeyError:
            pass


    def set(self, *args):

        path = ".".join(args[:-1])
        value = args[-1]

        self.__db.set_key(path, value)
        utils.request_call(self.update_observer, self.OBS_UPDATE, path, value)



    def get(self, *args):

        path = ".".join(args)
        value = self.__db.get_key(path, self.UNDEF)

        return value




_singleton = _ConfigManager()
def ConfigManager(): return _singleton



#
# Class for a generic comparator-sorted binary tree.
#
class BinTree(object):

    def __init__(self, value = None):

        self._parent = None
        self._left_child = None
        self._right_child = None
        self._value = value


    def _find_node(self, value, comparator):

        if (self._value == None):
            return (self, False)
        else:
            comp = comparator(value, self._value)
            if (comp < 0 and self._left_child):
                return self._left_child._find_node(value, comparator)
            elif (comp > 0 and self._right_child):
                return self._right_child._find_node(value, comparator)
            elif (comp == 0):
                return (self, True)
            else:
                return (self, False)


    def _get_successor(self):

        succ = self._right_child
        while (succ._left_child):
            succ = succ._left_child

        return succ


    def insert(self, value, comparator):

        node, success = self._find_node(value, comparator)

        if (node._value == None):
            node._value = value
        else:
            new_node = BinTree(value)
            new_node._parent = node
            if (comparator(value, node._value) < 0):
                node._left_child = new_node
            else:
                node._right_child = new_node


    def find(self, value, comparator):

        node, success = self._find_node(value, comparator)
        if (success): return node._value


    def remove(self, value, comparator):

        node, success = self._find_node(value, comparator)
        if (success):
            if (not node._left_child or not node._right_child):
                y = node
            else:
                y = node._get_successor()

            if (y._left_child):
                x = y._left_child
            else:
                x = y._right_child

            if (x):
                x._parent = y._parent

            if (not y._parent):
                if (x):
                    self._value = x._value
                else:
                    self._value = None
            else:
                if (y == y._parent._left_child):
                    y._parent._left_child = x
                else:
                    y._parent._right_child = x

            if (y != node):
                node._value = y._value




from ConfigWidget import ConfigWidget
from utils.datatypes import *

import gtk


class ConfigColor(ConfigWidget):

    def __init__(self, name, getter, setter, caller):

        ConfigWidget.__init__(self, name, getter, setter, caller)

        self._register_property("value", TYPE_STRING, self._setp_value,
                                self._getp, "black", doc = "Color value")


    def get_widgets(self):

        self.__label = gtk.Label("")
        self.__label.show()
        align = gtk.Alignment(0.0, 0.0, 0.0, 0.0)
        align.show()
        align.add(self.__label)

        self.__colorpicker = gtk.ColorButton()
        self.__colorpicker.set_use_alpha(True)
        self.__colorpicker.show()
        self.__colorpicker.connect("color-set", self.__on_change)

        return (align, self.__colorpicker)


    def _set_enabled(self, value):

        self.__colorpicker.set_sensitive(value)


    def _set_label(self, value):

        self.__label.set_text(value)



    def __on_change(self, src):

        color = src.get_color()
        alpha = src.get_alpha()
        value = "#%02X%02X%02X%02X" % (color.red >> 8, color.green >> 8,
                                       color.blue >> 8, alpha >> 8)
        self._set_config(value)


    def _setp_value(self, key, value):

        import utils
        r, g, b, alpha = utils.parse_color(value)
        self.__colorpicker.set_color(gtk.gdk.Color(r << 8, g << 8, b << 8))
        self.__colorpicker.set_alpha((alpha << 8) + alpha)
        self._set_config(value)
        self._setp(key, value)



import gtk


#
# Class for the configuration widgets of a sensor.
#
class SensorConfigurator(gtk.VBox):

    CHANGE_ENTRY = 0
    CHANGE_OPTION = 1
    CHANGE_SPIN = 2
    CHANGE_CHECKBOX = 3
    CHANGE_FONT = 4
    CHANGE_COLOR = 5
    CHANGE_FILE = 6


    def __init__(self, set_config, get_config):

        # functions for setting / getting configuration values of the sensor
        self.__set_config = set_config
        self.__get_config = get_config

        # name of the configurator
        self.__name = ""

        # the number of widget lines
        self.__lines = 0


        gtk.VBox.__init__(self)
        self.set_border_width(12)
        self.show()

        self.__table = gtk.Table(1, 2)
        self.__table.show()
        self.add(self.__table)

        self.__tooltips = gtk.Tooltips()

        self.callbacks = []


    #
    # Reacts on changing a setting.
    #
    def __on_change(self, src, *args):

        property, mode = args[-2:]
        args = args[:-2]

        if (mode == self.CHANGE_ENTRY):
            value = src.get_text()

        elif (mode == self.CHANGE_OPTION):
            items = args[0]
            value = items[src.get_active()]

        elif (mode == self.CHANGE_CHECKBOX):
            value = src.get_active()

        elif (mode == self.CHANGE_SPIN):
            value = src.get_value_as_int()

        elif (mode == self.CHANGE_FONT):
            value = src.get_font_name()

        elif (mode == self.CHANGE_COLOR):
            color = src.get_color()
            value = "#%02X%02X%02X" % (color.red >> 8, color.green >> 8,
                                       color.blue >> 8)

        self.__set_config(property, value)


    #
    # Sets/returns the name of this configurator. That name will appear in the
    # notebook tab.
    #
    def set_name(self, name): self.__name = name
    def get_name(self): return self.__name



    #
    # Adds a line of widgets to the configurator. The line can be indented.
    #
    def __add_line(self, indent, w1, w2 = None):

        self.__lines += 1
        self.__table.resize(self.__lines, 2)

        if (indent): x, y = 12, 3
        else: x, y = 0, 3

        if (w2):
            self.__table.attach(w1, 0, 1, self.__lines - 1, self.__lines,
                                gtk.FILL, 0, x, y)
            self.__table.attach(w2, 1, 2, self.__lines - 1, self.__lines,
                                gtk.EXPAND | gtk.FILL, 0, 0, y)

        else:
            self.__table.attach(w1, 0, 2, self.__lines - 1, self.__lines,
                                gtk.EXPAND | gtk.FILL, 0, x, y)



    def add_title(self, label):

        lbl = gtk.Label("")
        lbl.set_markup("<b>" + label + "</b>")
        lbl.show()
        align = gtk.Alignment()
        align.show()
        align.add(lbl)

        self.__add_line(0, align)



    def add_checkbox(self, label, property, help):

        check = gtk.CheckButton(label)
        check.show()

        self.__tooltips.set_tip(check, help)
        self.__add_line(1, check)

        value = self.__get_config(property)
        check.set_active(value)
        check.connect("toggled", self.__on_change, property,
                      self.CHANGE_CHECKBOX)



    def add_entry(self, label, property, help, passwd = 0):

        lbl = gtk.Label(label)
        lbl.show()
        entry = gtk.Entry()
        entry.show()
        align = gtk.Alignment()
        align.show()
        align.add(lbl)

        if (passwd):
            entry.set_visibility(False)
            entry.set_invisible_char(unichr(0x2022))

        self.__tooltips.set_tip(entry, help)
        self.__add_line(1, align, entry)

        value = self.__get_config(property)
        entry.set_text(value)
        entry.connect("changed", self.__on_change, property,
                      self.CHANGE_ENTRY)



    def add_spin(self, label, property, help, low, up):

        lbl = gtk.Label(label)
        lbl.show()

        align = gtk.Alignment()
        align.show()
        align.add(lbl)

        adjustment = gtk.Adjustment(0, int(low), int(up), 1, 1, 0)
        spin_button = gtk.SpinButton(adjustment, 1, 0)
        spin_button.set_numeric(True)
        spin_button.show()

        value = self.__get_config(property)

        self.__tooltips.set_tip(spin_button, help)
        self.__add_line(1, align, spin_button)

        spin_button.set_value(value)
        spin_button.connect("value-changed", self.__on_change, property,
                            self.CHANGE_SPIN)



    def add_option(self, label, property, help, options):

        lbl = gtk.Label(label)
        lbl.show()

        align = gtk.Alignment()
        align.show()
        align.add(lbl)

        value = self.__get_config(property)

        optmenu = gtk.combo_box_new_text()
        optmenu.show()
        items = []
        cnt = 0
        index = 0
        for k, v in options:
            optmenu.append_text(k)
            items.append(v)
            if (v == value): index = cnt
            cnt += 1
        #end for
        optmenu.set_active(index)
        optmenu.connect("changed", self.__on_change, items, property,
                        self.CHANGE_OPTION)

        self.__tooltips.set_tip(optmenu, help)
        self.__add_line(1, align, optmenu)



    def add_font_selector(self, label, property, help):

        lbl = gtk.Label(label)
        lbl.show()
        align = gtk.Alignment(0.0, 0.0, 0.0, 0.0)
        align.show()
        align.add(lbl)

        btn = gtk.FontButton()
        btn.set_use_font(True)
        btn.set_use_size(False)
        btn.set_show_style(True)
        btn.set_show_size(True)
        btn.show()
        btn.connect("font-set", self.__on_change, property,
                     self.CHANGE_FONT)

        font = self.__get_config(property)
        btn.set_font_name(font)

        self.__tooltips.set_tip(btn, help)
        self.__add_line(1, align, btn)



    def add_color_selector(self, label, property, help):

        lbl = gtk.Label(label)
        lbl.show()
        align = gtk.Alignment(0.0, 0.0, 0.0, 0.0)
        align.show()
        align.add(lbl)

        btn = gtk.ColorButton()
        btn.set_use_alpha(True)
        btn.show()
        print property
        btn.connect("color-set", self.__on_change, property, self.CHANGE_COLOR)

        colorname = self.__get_config(property)
        color = gtk.gdk.color_parse(colorname)

        btn.set_color(color)

        self.__tooltips.set_tip(btn, help)
        self.__add_line(1, align, btn)



    def add_file_selector(self, label, property, help):

        def open_filedialog(self, on_ok, entry):
            def fd_hide(src, fselector): fselector.destroy()
            d = gtk.FileSelection()
            d.show()
            d.ok_button.connect("clicked", on_ok, d, entry)
            d.cancel_button.connect("clicked", fd_hide, d)

        def on_ok(src, fselector, entry):
            fname = fselector.get_filename()
            entry.set_text(fname)
            fselector.destroy()

        lbl = gtk.Label(label)
        lbl.show()
        align = gtk.Alignment()
        align.show()
        align.add(lbl)

        hbox = gtk.HBox()
        hbox.show()
        entry = gtk.Entry()
        entry.show()
        hbox.pack_start(entry, True, True, 0)

        btn = gtk.Button(stock = gtk.STOCK_OPEN)
        btn.show()
        hbox.pack_end(btn, True, True, 4)

        btn.connect("clicked", open_filedialog, on_ok, entry)

        value = self.__get_config(property)
        entry.set_text(value)
        entry.connect("changed", self.__on_change, property,
                      self.CHANGE_ENTRY)

        self.__tooltips.set_tip(entry, help)
        self.__tooltips.set_tip(btn, help)
        self.__add_line(1, align, hbox)

    #
    # Adds a function to be called when the user closes the configurator
    #
    def add_close_callback(self, fct, args=None):

        self.callbacks.append( (fct, args) )



from ConfigWidget import ConfigWidget
from utils.datatypes import *

import gtk

class ConfigDPI(ConfigWidget):

    def __init__(self, name, getter, setter, caller):

        ConfigWidget.__init__(self, name, getter, setter, caller)

        self._register_property("value", TYPE_INT, self._setp_value,
                                self._getp, "", doc = "DPI value")


    def get_widgets(self):

        vbox = gtk.VBox(spacing = 3)
        vbox.show()

        hbox = gtk.HBox(spacing = 6)
        hbox.show()
        vbox.add(hbox)
        lbl = gtk.Label("dpi:")
        hbox.pack_start(lbl, False, False)

        spin = gtk.SpinButton()
        spin.show()
        spin.set_range(10, 500)
        spin.set_increments(1, 10)
        spin.set_digits(0)
        spin.connect("value-changed", self.__on_change)
        hbox.pack_start(spin, False, False)

        self.__line = gtk.Viewport()
        self.__line.show()
        self.__line.modify_bg(gtk.STATE_NORMAL, gtk.gdk.color_parse("grey80"))
        hbox.pack_start(self.__line, False, False)

        lbl = gtk.Label("<--  5 cm  /  1.97\"  -->")
        lbl.show()
        lbl.modify_fg(gtk.STATE_NORMAL, gtk.gdk.color_parse("black"))
        self.__line.add(lbl)

        self.__label = gtk.Label("")
        self.__label.show()
        vbox.add(self.__label)

        self.__box = vbox
        self.__spin = spin
        return (vbox,)


    def __on_change(self, src, *args):

        value = int(src.get_value())
        self.__set_line(value)
        self._set_config(value)


    def __set_line(self, dpi):

        width = int(5 * (1 / 2.54) * dpi)
        self.__line.set_size_request(width, 10)


    def _set_enabled(self, value): self.__box.set_sensitive(value)
    def _set_label(self, value): self.__label.set_markup(value)



    def _setp_value(self, key, value):

        self.__spin.set_value(value)
        self._setp(key, value)



try:
    set
except NameError:
    from sets import Set as set

#
# Class for parsing expressions in a simple query language.
#
# Grammar:  CLAUSE -> ( CMD ARGS )
#           ARGS ->   ARG ARGS | e
#           ARG ->    CLAUSE | STRING
#           STRING -> ".*" | '.*'
#           CMD ->    [A-Z]+
#
# Usage:    myparser = QueryParser(mytrie)
#           result = myparser.parse(query)
#
class QueryParser(object):

    __slots__ = '__trie'

    def __init__(self, trie):

        self.__trie = trie


    def parse(self, query):

        query, result = self.__parse_CLAUSE(query)
        if (query.strip()): raise SyntaxError("trailing garbage: %s" % (query))

        return result


    def __parse_CLAUSE(self, query):

        query = query.lstrip()
        token = query[0]
        query = query[1:]
        if (token != "("): raise SyntaxError("'(' expected.")

        # read command
        query = query.lstrip()
        pos = query.find(" ")
        cmd = query[:pos]
        query = query[pos:]

        if (cmd == "AND"):
            query, r1 = self.__parse_ARG(query)
            query, r2 = self.__parse_ARG(query)
            result = set(r1)
            result.intersection_update(r2)
            result = list(result)

        elif (cmd == "OR"):
            query, r1 = self.__parse_ARG(query)
            query, r2 = self.__parse_ARG(query)
            result = set(r1)
            result.union_update(r2)
            result = list(result)

        elif (cmd == "MATCH"):
            query, r1 = self.__parse_ARG(query)
            query, r2 = self.__parse_ARG(query)
            result = self.__trie.retrieve([r1] + list(r2))

        # read paren
        query = query.lstrip()
        token = query[0]
        if (token != ")"): raise SyntaxError("')' expected. " + query)
        query = query[1:]

        return (query, result)


    def __parse_ARG(self, query):

        query = query.lstrip()
        token = query[0]

        if (token == "("):
            query, result = self.__parse_CLAUSE(query)

        elif (token in ["\"", "'"]):
            query, result = self.__parse_STRING(query)

        else:
            raise SyntaxError("'(' or '\"' expected.")

        return (query, result)


    def __parse_STRING(self, query):

        delimiter = query[0]
        value = ""
        cnt = 1
        c = query[cnt]
        while (c != delimiter):
            value += c
            cnt += 1
            c = query[cnt]
        #end while

        query = query[cnt + 1:]
        return (query, value)



import os
import commands
import tempfile


_BZIP2 = 0
_GZIP = 1


#
# Class for accessing desklet packages.
#
class Package:

    def __init__(self, file):

        file = os.path.abspath(file)
        self.__dir = tempfile.mktemp()
        os.mkdir(self.__dir)

        # determine archive type
        head = open(file, "r").read(3)
        if (head.startswith("BZh")):
            cmd = "cd \"%s\" && tar -xjf \"%s\"" % (self.__dir, file)
        elif (head.startswith("\x1f\x8b")):
            cmd = "cd \"%s\" && tar -xzf \"%s\"" % (self.__dir, file)
        else:
            raise IOError("Unknown file type.")

        fail, out = commands.getstatusoutput(cmd)
        if (fail): raise IOError(out)

        

    def close(self):

        os.system("rm -rf \"%s\"" % (self.__dir))


    def find_displays(self):

        def f(file): return (os.path.splitext(file)[1] == ".display")
        values = self.__find(self.__dir, f)
        return [os.path.dirname(p) for p in values]



    def find_controls(self):

        def f(file): return (file == "controls")
        return self.__find(self.__dir, f)


    def find_sensors(self):

        def f(file): return (file[:7] == "Install" and
                             os.path.splitext(file)[1] == ".bin")
        return self.__find(self.__dir, f)



    def __find(self, path, predicate):

        if (not path): path = self.__dir
        values = []
        files = os.listdir(path)
        for f in files:
            if (predicate(f)):
                values.append(os.path.join(path, f))
            elif (os.path.isdir(os.path.join(path, f))):
                values += self.__find(os.path.join(path, f), predicate)
        #end for
        
        return values


    def install_displays(self, dest):

        pass



    def install_controls(self, dest):

        pass


    def install_sensors(self, dest):

        pass



from config.StateSaver import DefaultStateSaver


#
# Abstract base class for plugins.
#
class Plugin(object):

    def init(self): pass


    def _get_path(self): return self._path


    def _get_plugins_by_pattern(self, key, pattern):

        from PluginRegistry import registry
        return registry.get_plugins_by_pattern(key, pattern)


    def _get_plugin(self, name, *interfaces):

        from PluginRegistry import registry
        plugin = registry.get_plugin(name, *interfaces)
        if (not plugin):
            print "Plugin not found:", name

        return plugin


    def _set_config(self, key, value):

        backend = DefaultStateSaver()
        backend.set_key(key, value)


    def _get_config(self, key):

        backend = DefaultStateSaver()
        return backend.get_key(key, True)



__all__ = ('READWRITE', 'READ', 'WRITE', 'DENY', 'Permission')


class _Permission(object):

    READ  = 0x4
    WRITE = 0x2

    __slots__ = ('__mode',)


    def __init__(self, prop):

        self.__mode = 0x0

        if prop.fget:
            self.__mode |= _Permission.READ

        if prop.fset:
            self.__mode |= _Permission.WRITE



    def __get_readable(self):
        return self.__mode & _Permission.READ

    def __get_writable(self):
        return self.__mode & _Permission.WRITE

    Readable = property(__get_readable)
    Writable = property(__get_writable)



    def __str__(self):

        s = ''

        if self.Readable:
            s += 'r'
        if self.Writable:
            s += 'w'

        return s



    def __int__(self):

        return self.__mode




class __Foo(object):

    read  = property(fget=int)
    write = property(fset=int)
    rw    = property(fget=int, fset=int)
    deny  = property(fget=None, fset=None)




READ      = _Permission(__Foo.read)
WRITE     = _Permission(__Foo.write)
READWRITE = _Permission(__Foo.rw)
DENY      = _Permission(__Foo.deny)




def Permission(prop = None):

    if not prop:
        return DENY

    if prop.fget and prop.fset:
        return READWRITE

    elif prop.fget:
        return READ

    elif prop.fset:
        return WRITE



from ContainerTarget import ContainerTarget
#from utils.TilingImage import TilingImage
from utils.tiling import Tiling as TilingImage
from utils.datatypes import *
from layout import Unit
from utils import vfs
import utils

import os
import gtk


#
# Class for putting frames around targets.
#
class TargetFrame(ContainerTarget):

    def __init__(self, name, parent):

        self.__size = (0, 0)
        self.__inner_size = (0, 0)
        self.__frame_widths = [Unit.ZERO] * 4
        self.__images = []


        ContainerTarget.__init__(self, name, parent)
        self.__table = gtk.Table(3, 3)
        self.__table.set_direction(gtk.TEXT_DIR_LTR)
        self.__table.show()

        r, g, b, a = utils.parse_color("black")
        for x, y in [(0, 1), (1, 0), (2, 1), (1, 2),
                     (0, 0), (2, 0), (2, 2), (0, 2)]:
            img = TilingImage()
            img.set_from_color(r, g, b, a)
            img.show()
            self.__table.attach(img, x, x + 1, y, y + 1)
            self.__images.append(img)

        self.__box = gtk.HBox()
        self.__box.show()
        self.__table.attach(self.__box, 1, 2, 1, 2)

        self._register_property("border-uris", TYPE_LIST,
                                self._setp_border_uris, self._getp)
        self._register_property("border-width", TYPE_UNIT_LIST,
                                self._setp_border_width, self._getp)
        self._register_property("color", TYPE_STRING,
                                self._setp_border_color, self._getp)

        self.set_prop("border-width",
                      [Unit.Unit(2, Unit.UNIT_PX), Unit.Unit(2, Unit.UNIT_PX),
                       Unit.Unit(2, Unit.UNIT_PX), Unit.Unit(2, Unit.UNIT_PX)])
        self._setp("color", "black")


        # watch for geometry changes
        self.add_observer(self.__on_observe_size)



    def get_widget(self): return self.__table

    def new_child(self, childtype, settings, children):

        child = ContainerTarget.new_child(self, childtype, settings, children)
        self.__box.add(child.get_widget())
        self.__redraw_frame()



    #
    # Observer for size changes.
    #
    def __on_observe_size(self, src, cmd, *args):

        x, y, w, h = src.get_geometry()
        if (cmd == src.OBS_GEOMETRY and
            (w.as_px(), h.as_px()) != self.__size):
            self.__redraw_frame()



    def __redraw_frame(self, src = None, event = None):

        x, y, w, h = self.get_geometry()
        w1, h1, w2, h2 = self.__frame_widths

        iw = max(Unit.ZERO, w - w1 - w2)
        ih = max(Unit.ZERO, h - h1 - h2)

        self.__box.set_size_request(iw.as_px(), ih.as_px())
        cnt = 0
        for bw, bh in ((w1, ih), (iw, h1),
                       (w2, ih), (iw, h2),
                       (w1, h1), (w2, h1),
                       (w2, h2), (w1, h2)):
            img = self.__images[cnt]
            cnt += 1

            if (bw.as_px() == 0 or bh.as_px() == 0):
                img.hide()
            else:
                img.show()

                img.tile(bw.as_px(), bh.as_px())
                img.set_size_request(bw.as_px(), bh.as_px())
        #end for



    def __set_border_width(self, args):

        for i in xrange(len(args)):
            self.__frame_widths[i] = args[i]

        self.__size = (0, 0)

        w1, h1, w2, h2 = self.__frame_widths
        self.get_layout_object().set_border_width(*args)



    def __set_border(self, args):

        cnt = 0
        for uri in args:
            uri = self._get_display().get_full_path(uri)
            if (vfs.exists(uri)):
                try:
                    data = vfs.read_entire_file(uri)
                except:
                    return
                self.__images[cnt].set_from_data(data)
            cnt += 1

        self.__size = (0, 0)
        self.__redraw_frame()



    def __set_border_color(self, color):

        r, b, g, a = utils.parse_color(color)
        for cnt in range(8):
            self.__images[cnt].set_from_color(r, g, b, a)

        self.__size = (0, 0)
        self.__redraw_frame()



    def get_border_size(self):

        w1, h1, w2, h2 = self.__frame_widths
        return (w1, h1, w2, h2)



    def _setp_border_uris(self, key, value):

        self.__set_border(value)
        self._setp(key, value)



    def _setp_border_color(self, key, value):

        self.__set_border_color(value)
        self._setp(key, value)



    def _setp_border_width(self, key, value):

        self.__set_border_width(value)
        self._setp(key, value)



from DisplayTarget import DisplayTarget
from utils.datatypes import *
from layout import Unit
from utils.Struct import Struct

import gobject
import gtk
import pango
import re


class TargetEntry(DisplayTarget):

    __ACTION_CHANGE = "change"

    # regular expression for parsing font description strings
    __RE_FONT = re.compile("(?P<name>.+?)"
                           "(?P<size>[0-9\.]+)"
                           "(?P<unit>[a-z]+)?$")


    def __init__(self, name, parent):

        DisplayTarget.__init__(self, name, parent)

        self.__widget = gtk.Entry()
        self.__widget.show()
        self.__widget.connect("changed", self.__on_changed)

        self._register_property("value", TYPE_STRING,
                                self._setp_value, self._getp)
        self._register_property("font", TYPE_STRING,
                                self._setp_font, self._getp)
        self._register_property("color", TYPE_STRING,
                                self._setp_color, self._getp)

        self._register_action(self.__ACTION_CHANGE)

        w, h = self.__widget.size_request()
        self.set_size(Unit.Unit(w, Unit.UNIT_PX), Unit.Unit(h, Unit.UNIT_PX))



    def get_widget(self): return self.__widget



    def __on_changed(self, src):

        self._setp("value", self.__widget.get_text())
        self._get_display().send_action(self, self.__ACTION_CHANGE, Struct())



    def _setp_font(self, key, value):

        if (not self.__widget.window):
            gobject.idle_add(self._setp_font, key, value)
            return

        m = re.match(self.__RE_FONT, value)
        if (m):
            name = m.group("name")
            unit = m.group("unit") or Unit.UNIT_PT
            size = m.group("size")
            if (size):
                unit = Unit.Unit(float(size), unit)
                size = unit.as_pt()
                value = "%s %f" % (name, size)

        fd = pango.FontDescription(value)
        self.__widget.modify_font(fd)

        # We need this hack to ensure that GTK resizes the widget after
        # having changed the font size.
        def f(self):
            w, h = self.__widget.size_request()
            self.set_size(Unit.Unit(w, Unit.UNIT_PX),
                          Unit.Unit(h, Unit.UNIT_PX))

        self.__widget.set_size_request(-1, -1)
        gobject.idle_add(f, self)
        self._setp(key, value)


    def _setp_color(self, key, value):

        self.__widget.modify_text(gtk.STATE_NORMAL, gtk.gdk.color_parse(value))

        self._setp(key, value)


    def _setp_value(self, key, value):

        self.__widget.set_text(value)



from libdesklets.controls import Interface, Permission

class ISystem(Interface):

    allfs            = Permission.READWRITE
    iface            = Permission.READWRITE
    mountdir         = Permission.READWRITE
    pid              = Permission.READWRITE
    which            = Permission.READWRITE

    cpu              = Permission.READ
    cpu_bogomips     = Permission.READ
    cpu_cache        = Permission.READ
    cpu_clock        = Permission.READ
    cpu_load         = Permission.READ
    cpu_model        = Permission.READ

    fsusage          = Permission.READ
    
    hostname         = Permission.READ    
    kernel_version   = Permission.READ
    loadavg          = Permission.READ
    memory           = Permission.READ
    mountlist        = Permission.READ
    net_devices      = Permission.READ
    net_load         = Permission.READ
    net_speed        = Permission.READ
    net_state        = Permission.READ

    operating_system = Permission.READ
    ppp              = Permission.READ
    proc_args        = Permission.READ
    proc_kernel      = Permission.READ
    proc_list        = Permission.READ
    proc_map         = Permission.READ
    proc_mem         = Permission.READ
    proc_segment     = Permission.READ
    proc_signal      = Permission.READ
    proc_state       = Permission.READ
    proc_time        = Permission.READ
    proc_uid         = Permission.READ
    swap             = Permission.READ
    uptime           = Permission.READ
    users            = Permission.READ




from ConfigWidget import ConfigWidget
from utils.datatypes import *
from utils import vfs

import gtk
import os

class ConfigURI(ConfigWidget):

    def __init__(self, name, getter, setter, caller):

        ConfigWidget.__init__(self, name, getter, setter, caller)

        self._register_property("value", TYPE_STRING, self._setp_value,
                                self._getp, "", doc = "Value")



    def get_widgets(self):

        def open_filedialog(src, self, entry):
            def response_cb(src, response):
                if (response == gtk.RESPONSE_OK):
                    fname = src.get_filename()
                    if (fname):
                        entry.set_text(fname)
                        self.__on_change(entry, None)
                    src.destroy()
                else:
                    src.destroy()

            def preview_cb(src):
                fname = src.get_preview_filename()
                success = src.get_preview_widget().preview(fname)
                d.set_preview_widget_active(success)


            preview = _FileChooserPreview()
            preview.show()
            d = gtk.FileChooserDialog("", None,
                                      gtk.FILE_CHOOSER_ACTION_OPEN,
                                      (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL,
                                       gtk.STOCK_OPEN, gtk.RESPONSE_OK))
            d.set_local_only(False)
            path = os.path.join(self._get_path(), entry.get_text())
            d.set_current_folder_uri(os.path.dirname(path))
            d.set_preview_widget(preview)
            d.set_preview_widget_active(False)
            d.set_use_preview_label(False)
            d.show()
            d.connect("response", response_cb)
            d.connect("selection-changed", preview_cb)


        self.__label = gtk.Label("")
        self.__label.show()
        align = gtk.Alignment(0.0, 0.0, 0.0, 0.0)
        align.show()
        align.add(self.__label)

        hbox = gtk.HBox()
        hbox.show()
        self.__entry = gtk.Entry()
        self.__entry.show()
        hbox.pack_start(self.__entry, True, True, 0)

        btn = gtk.Button(stock = gtk.STOCK_OPEN)
        btn.show()
        hbox.pack_end(btn, True, True, 4)

        btn.connect("clicked", open_filedialog, self, self.__entry)
        self.__entry.connect("focus-out-event", self.__on_change)

        # we need an EventBox for being able to display tooltips
        self.__ebox = gtk.EventBox()
        self.__ebox.show()
        self.__ebox.add(hbox)

        return (align, self.__ebox)


    def __on_change(self, src, event):

        value = src.get_text()
        self._set_config(value)


    def _set_label(self, value): self.__label.set_text(value)
    def _set_enabled(self, value): self.__ebox.set_sensitive(value)


    def _setp_value(self, key, value):

        self.__entry.set_text(value)
        self._set_config(value)
        self._setp(key, value)



#
# Preview widget for previewing files. Returns whether the given file could be
# displayed.
#
class _FileChooserPreview(gtk.Frame):

    def __init__(self):

        gtk.Frame.__init__(self)
        self.set_shadow_type(gtk.SHADOW_IN)
        self.__img = gtk.Image()
        self.__img.show()
        self.add(self.__img)


    def preview(self, filename):

        loader = gtk.gdk.PixbufLoader()
        try:
            fd = vfs.open(filename)
            data = fd.read(3 * 1024 * 1024)  # read a maximum of 3 MB
            fd.close()
            loader.write(data, len(data))

        except:
            try:
                loader.close()
            except:
                pass
            return False

        try:
            loader.close()
        except:
            return False

        pbuf = loader.get_pixbuf()
        if (pbuf == None):
            return False

        # scale image down while preserving aspect ratio
        width = pbuf.get_width()
        height = pbuf.get_height()
        if (width > 180):
            scale = 180 / float(width)
            height *= scale
            height = max(1, height)
            if (abs(scale - 1.0) > 0.001):
                pbuf = pbuf.scale_simple(180, int(height), 3)

        self.__img.set_size_request(180, -1)
        self.__img.set_from_pixbuf(pbuf)
        self.__img.set_size_request(180, -1)
        del pbuf
        import gc; gc.collect()

        return True



TIPS = (
    _("You can find lots of desktop applets for use\n"
      "with gDesklets at\n"
      "\n"
      "http://gdesklets.gnomedesktop.org"),

    _("You can install new desktop applets easily by\n"
      "dragging the download link for the package\n"
      "from your web browser into the gDesklets Shell."),

    _("Click the right mouse button on a desktop\n"
      "applet or press F10 to open the popup menu."),

    _("You can move the applets around the screen\n"
      "by holding down the middle mouse button.\n"
      "\n"
      "If your mouse has only two buttons,\n"
      "hold down both of them simultaneously\n"
      "in order to simulate the middle button."),

    _("You can have different profiles of\n"
      "desktop applets and switch between them.\n"
      "\n"
      "Select \"Profiles->New profile...\" from\n"
      "the menu bar to create new profiles.\n"
      "\n"
      "Empty profiles get discarded automatically."),

    _("The search bar accepts complex search\n"
      "expressions for finding your applets."),

    _("Search expressions may contain wildcards:\n"
      "\n"
      " ? matches any one character\n"
      " * matches zero or more characters")
    )



COMMAND_OPEN_DISPLAY = "open display"
COMMAND_OPEN_DISPLAY_WITH_ID = "open display with id"
COMMAND_OPEN_PLUG = "open plug"
COMMAND_CLOSE_DISPLAY = "close display"
COMMAND_SHUTDOWN = "shutdown"
COMMAND_DISPLAYS = "displays"
COMMAND_DISPLAY_LIST = "display list"
COMMAND_GEOMETRY = "geometry"
COMMAND_SET_REMOVE_COMMAND = "set remove command"
COMMAND_VERSION = "version"
COMMAND_ABOUT = "about"
COMMAND_CONFIGURE = "configure"



#:section Sensor
#
#:  para The [cmd:Sensor] class is an abstract base class where new sensors are
#        derived from. This class takes care of communication with the
#        [app:gDesklets] core, so you can fully concentrate on the sensor
#        itself.

from utils.Observable import Observable
from utils.TargetSettings import TargetSettings
from utils import i18n
from utils import MAIN_THREAD
from utils import run_in_main_thread
from config.ConfigManager import ConfigManager
from SensorConfigurator import SensorConfigurator
from main import _
from main import PURGE_KEY

import gobject
import gtk
import thread
import time
import os


#
#:class Sensor |
# Abstract base class for all sensors. Every sensor has to inherit from this
# class.
#
class Sensor(Observable):

    OBS_OUTPUT = 0
    OBS_CMD_CONFIGURE = 1
    OBS_CMD_REMOVE = 2
    OBS_CMD_RESTART = 3
    OBS_CMD_DUPLICATE = 4
    OBS_CMD_MENU = 5


    def __init__(self):

        # the path of this sensor
        self.__path = os.getcwd()

        # configuration manager for sensor properties
        self.__config_manager = ConfigManager()
        self.__config_id = ""

        # the ID of this sensor
        self.__sensor_id = ""

        # the configuration watcher callback
        self.__config_watcher = None

        # flag for stopping the sensor
        self.__stop_flag = False

        # when stopping, the flag is set to True, then the callback is called
        self.__stop_callback = Null

        self.__defaults = {}



    #
    # Sets the ID of this sensor.
    #
    def set_id(self, ident):

        self.__id = ident



    #
    #:function get_id | |
    #          Returns the ID of this sensor. It is not yet valid in the
    #          constructor.
    #:  return string
    #:/function
    #
    def get_id(self):
        assert(self.__id, "The ID is invalid in the constructor.")

        return self.__id



    #
    # Sets the configuration ID for this sensor.
    # If the user has set a callback function for watching the config,
    # the config will be watched.
    #
    def set_config_id(self, config_id):

        self.__config_id = str(config_id)
        if (self.__config_watcher):
            Sensor.watch_config(self, self.__config_watcher)



    #
    #:function get_config_id | |
    #          Returns the unique config ID of the sensor. You can use the ID
    #          if you need to store data in other places.
    #:  return string
    #:/function
    #
    def get_config_id(self):

        return self.__config_id



    #
    #:function set_config | key, value |
    #          Stores the given value in the configuration base.
    #:  param key   | string | The key name.
    #:  param value | string | The value to set.
    #:/function
    #
    def set_config(self, key, value):
        assert(self.__config_manager)

        if (not self.__stop_flag):
            run_in_main_thread(self.__config_manager.set,
                               self.__config_id, key, value)



    #
    #:function get_config | key |
    #          Returns the configuration value for the given key.
    #:  param  key | string | The key name.
    #:  return string
    #:/function
    #
    def get_config(self, key):
        assert(self.__config_manager)

        value = run_in_main_thread(self.__config_manager.get,
                                   self.__config_id, key)

        if (value == self.__config_manager.UNDEF):
            value = self.__defaults.get(key, "")

        return value



    #
    #:function watch_config | callback |
    #          Registers a watcher for config changes.
    #:  param callback | function | The callback function for configuration
    #                               changes.
    #:/function
    #
    def watch_config(self, callback):

        self.__config_watcher = callback
        if (self.get_config_id()):
            self.__config_manager.watch(self.get_config_id(),
                                        self.__on_watch_config)



    #
    #:function _set_config_type | key, type, default |
    #          Sets the data types to be used for the configuration values.
    #          Use this method in the constructor of your sensor.
    #:  param  key | string | The name of the configuration key.
    #:  param  type | enum  | The data type of the configuration key.
    #:  param  default | string | The default value or unset keys.
    #:/function
    #
    def _set_config_type(self, key, dtype, default):

        self.__defaults[key] = default


    #
    #:function get_path | | Returns the filesystem path of the sensor. Use this
    #                       method if you want to load resource files that come
    #                       with the sensor.
    #:  return string
    #:/function
    #
    def get_path(self): return self.__path



    #
    #:function set_path_to_purge | paths |
    #          Sets the paths which are to be removed when the desklet that
    #          uses the sensor gets removed.
    #          If your sensor creates files, you need to specify the paths here
    #          in order to clean up.
    #:  param  paths | string list | The paths to purge.
    #:/function
    #
    def set_paths_to_purge(self, paths):

        key = PURGE_KEY
        value = ",".join(paths)
        self.__config_manager.set(self.__config_id,
                                  key, value)



    #
    #:function new_output || Returns a new empty [cmd:TargetSettings] object for sending data to
    #          the display.
    #:  return TargetSettings
    #:/function
    #
    def new_output(self): return TargetSettings()



    #
    #:function send_output | output | Sends the given [cmd:TargetSettings] object to the display.
    #          [emph:Never call this method from within a thread!]
    #:  param  output | TargetSettings | The object for sending to the display.
    #:/function
    #
    def send_output(self, output):

        run_in_main_thread(self.update_observer, self.OBS_OUTPUT, output)



    #
    #:function add_timer | interval, callback, *args | Adds a timeout function with the given
    #          interval in ms.
    #:  param  interval | int | The timeout interval between each invokation of the callback.
    #:  param  callback | function | The callback function.
    #:/function
    #
    def add_timer(self, interval, callback, *args):

        def __wrap(self, callback, args):
            if (not self.__stop_flag):
                try:
                    return callback(*args)
                except:
                    from utils.error import Error
                    Error().handle("unknown")
                    return False
            else:
                return False

        return gobject.timeout_add(interval, __wrap, self, callback, args)



    #
    #:function add_thread | threadfunction, *args | Adds and runs a new thread.
    #          Use this to start new threads. It's recommended to put blocking actions into threads
    #          in order to not block [app:gDesklets].
    #:  param threadfunction | function | The thread function.
    #:/function
    #
    def add_thread(self, threadfunction, *args):

        # the thread should not start before setup is complete, therefore
        # we are using the GTK idle handler
        def run_thread(threadfunction, args):
            thread.start_new_thread(threadfunction, args)
            return False

        gobject.idle_add(run_thread, threadfunction, args)



    #
    # Sends an action to this sensor.
    #
    def send_action(self, call, path, args = []):

        if (not self.__stop_flag):
            try:
                self.call_action(call, path, args)
            except:
                from utils.error import Error
                Error().handle("unknown")




    #
    #:function call_function | call, path, *args | Method for handling action calls from the
    #          display. Sensors have to override this method. [emph:This method may soon be
    #          deprecated!]
    #:  param  call | string | The function to call.
    #:  param  path | int list | The path of the target on which the action occured.
    #:  param  args | any list | The list of arguments for the function call.
    #:/function
    #
    def call_action(self, call, path, args = []): raise NotImplementedError



    def watch_stop(self, callback):

        """
        Registers a callback that will be called when sensor will be stopped.

        @param callback:
        @type  callback: callable(callback)
        """

        self.__stop_callback = callback


    #
    # Stops this sensor and cleans up.
    #
    def stop(self):

        self.__stop_flag = True
        self.__stop_callback()

        del self.__config_watcher
        self.__config_manager.remove_watcher(self.get_config_id(),
                                             self.__on_watch_config)

        # shut down sensor
        self._shutdown()



    #
    #:function is_stopped | | Returns whether this sensor has been stopped. Use this method to
    #          check if your threads have to terminate.
    #:  return bool
    #:/function
    #
    def is_stopped(self): return self.__stop_flag



    #
    #:function _shutdown | | Executes tasks for sensor shutdown. Override this method if your
    #          sensor has to clean up things after it has been terminated.
    #:/function
    #
    def _shutdown(self): pass



    #
    #:function _new_configurator | | Creates and returns an empty new sensor configurator.
    #          [emph:This method may soon be deprecated!]
    #:  return SensorConfigurator.
    #:/function
    #
    def _new_configurator(self): return SensorConfigurator(self.set_config,
                                                           self.get_config)


    #
    # Returns the configurator of this sensor.
    #
    def get_configurator(self): pass



    #
    # Method for duplicating the display using this sensor
    #
    def _duplicate_display(self):

        self.update_observer(self.OBS_CMD_DUPLICATE)



    #
    #:function open_menu | menu | Opens the given popup menu. The menu is a list of
    #          [cmd:(label:str, sensitive:bool, submenu:list, callback:function, args: list)]
    #          tuples.
    #:  param  menu | list | The menu to open.
    #:/function
    #
    def open_menu(self, menu):

        menu += [()]
        run_in_main_thread(self.update_observer, self.OBS_CMD_MENU, menu)



    #
    # Reacts on changes in the configuration and delegates the call to the
    # user's callback.
    #
    def __on_watch_config(self, path, value):
        assert(self.__config_watcher)

        key = path[-1]
        self.__config_watcher(key, value)



    def __on_configure(self, src):

        self.update_observer(self.OBS_CMD_CONFIGURE, None)



    def __on_remove(self, src):

        self.update_observer(self.OBS_CMD_REMOVE, None)



    def __on_restart(self, src):

        self.update_observer(self.OBS_CMD_RESTART, None)



    # some required aliases for older sensors to work
    _get_id = get_id
    _get_path = get_path
    _new_output = new_output
    _send_output = send_output
    _add_timer = add_timer
    _add_thread = add_thread
    _is_stopped = is_stopped
    _get_config_id = get_config_id
    _watch_config = watch_config
    _set_config = set_config
    _get_config = get_config
    _set_paths_to_purge = set_paths_to_purge
    _open_menu = open_menu




from config import settings
from config.DaemonConfigger import DaemonConfigger
from display.Display import Display
from display.Window import Window
from display.Plug import Plug
from factory.DisplayFactory import DisplayFactory
from main import COPYRIGHT, HOME, NAME, USERHOME, VERSION
from main.AboutDialog import AboutDialog
from main.remotecommands import *
from main.RemoteSocket import RemoteSocket
from utils import dialog
from utils import vfs
from utils.error import Error

import gobject
import gtk
import os
import sys


# This class starts, restarts and kills Displays
class Starter:

    __DISPLAY_TYPE_WINDOW = "window"
    __DISPLAY_TYPE_PLUG = "plug"


    def __init__(self):

        # feedback dialog
        self.__feedback = LoadingFeedback()

        # the communication socket
        self.__socket = RemoteSocket()

        # the factory which reads the display's XML
        self.__factory = DisplayFactory()

        # the About dialog
        self.__about_dialog = AboutDialog(os.path.join(HOME, "data"))

        # the configuration dialog
        self.__configger = DaemonConfigger()

        # the command which gets called when the user removes a display
        self.__remove_command = ""

        # the set of open displays as a hashtable "id -> display"
        self.__open_displays = {}
        # the paths of the display files "id -> path"
        self.__display_paths = {}

        # set the message handlers for the socket
        self.__socket.add_message_handler(COMMAND_OPEN_DISPLAY,
                                          self.__handle_open_display)
        self.__socket.add_message_handler(COMMAND_OPEN_DISPLAY_WITH_ID,
                                          self.__handle_open_display_with_id)
        self.__socket.add_message_handler(COMMAND_OPEN_PLUG,
                                          self.__handle_open_plug)
        self.__socket.add_message_handler(COMMAND_CLOSE_DISPLAY,
                                          self.__handle_close_display)
        self.__socket.add_message_handler(COMMAND_SHUTDOWN,
                                          self.__handle_shutdown)
        self.__socket.add_message_handler(COMMAND_DISPLAYS,
                                          self.__handle_displays)
        self.__socket.add_message_handler(COMMAND_DISPLAY_LIST,
                                          self.__handle_display_list)
        self.__socket.add_message_handler(COMMAND_GEOMETRY,
                                          self.__handle_geometry)
        self.__socket.add_message_handler(COMMAND_VERSION,
                                          self.__handle_version)
        self.__socket.add_message_handler(COMMAND_ABOUT,
                                          self.__handle_about)
        self.__socket.add_message_handler(COMMAND_CONFIGURE,
                                          self.__handle_configger)
        self.__socket.add_message_handler(COMMAND_SET_REMOVE_COMMAND,
                                          self.__handle_set_remove_command)


        # socket ready, start handling requests
        self.__socket.start()

        # setup a nice systray icon
        if (not HAVE_WIN32 and settings.show_tray_icon):
            from main.TrayIcon import TrayIcon
            trayicon = TrayIcon()
            trayicon.set_menu([(None, _("_Manage desklets"),
                                self.__handle_manage),
                               (),
                               (gtk.STOCK_PROPERTIES, _("_Configuration"),
                                 self.__handle_configger),
                               (None, _("_View log"),
                                self.__handle_show_log),
                                (),
                               (None, _("_About..."),
                                self.__handle_about_dialog),
                               (),
                               (gtk.STOCK_QUIT, _("_Stop daemon"),
                                self.__handle_shutdown)])



    #
    # Reacts on observer messages from the display.
    #
    def __on_display_action(self, src, cmd, *args):

        if (cmd == src.OBS_CLOSE):
            ident = args[0]
            self.__close_display(ident)
            del src

        elif (cmd == src.OBS_RESTART):
            ident = args[0]
            xcoord, ycoord = src.get_geometry()[0:2]
            xcoord = xcoord.as_px()
            ycoord = ycoord.as_px()
            path = self.__display_paths[ident]
            self.__remove_display(ident)
            def f(*args): pass
            gobject.timeout_add(250, self.__handle_open_display_with_id,
                                f, path, ident)
            del src



    #
    # Adds the given display.
    #
    def __add_display(self, ident, path, displaytype):

        dsp = None
        container = None

        # can this happen?
        if (ident in self.__open_displays): return None

        
        Error().register(ident, path)
        
        # create display
        try:
            dsp = self.__create_display(ident, path)
        except:
            Error().handle(ident)
            #log("Warning: Couldn't add desklet \"%s\".\n%s"  % (path, exc))
            self.__close_display(ident)
            return None
        
        self.__open_displays[ident] = dsp
        self.__display_paths[ident] = path

        log("Adding \"%s\" with ID \"%s\" to the desklet list."
            % (path, ident))
        dsp.add_observer(self.__on_display_action)

        # create container widget
        if (displaytype == self.__DISPLAY_TYPE_WINDOW):
            container = Window(dsp)

        #elif (displaytype == self.__DISPLAY_TYPE_PLUG):
        #    container = Plug()
        #    dsp.set_container(container)
        #    container.show()
            
        # initialize display
        try:
            dsp.initialize()
        except:
            Error().handle(ident)
            self.__close_display(ident)
            return None

        return container



    #
    # Creates and returns a new display from the given data.
    #
    def __create_display(self, ident, path):

        try:
            data = vfs.read_entire_file(path)

        except:
            log("Could not open desklet file \"%s\"." % (path,))
            raise UserError(_("Could not open desklet file \"%s\"") % (path,),
                           _("The desklet file could not be opened because "
                             "the file was not readable."),
                            show_details = False)

        display = self.__factory.create_display(ident, data, path)

        return display



    #
    # Removes the given display.
    #
    def __remove_display(self, ident):

        try:
            display = self.__open_displays[ident]
        except KeyError, exc:
            log("Warning: Couldn't remove desklet with ID \"%s\".\n%s" \
                % (ident, exc))
            return

        log("Removing \"%s\" with ID \"%s\" from the desklet list."
            % (self.__display_paths[ident], ident))
        display.remove_display()
        Error().forget(ident)

        try:
            del self.__open_displays[ident]
            del self.__display_paths[ident]
        except StandardError:
            pass

        import gc
        gc.collect()


    #
    # Closes the given display.
    #
    def __close_display(self, ident):

        if (ident in self.__open_displays):
            display = self.__open_displays[ident]
            display.purge_display()

        self.__remove_display(ident)
        if (self.__remove_command):
            os.system("%s %s &" % (self.__remove_command, ident))



    def __handle_open_display(self, callback, path):

        ident = Display.make_id()
        self.__handle_open_display_with_id(callback, path, ident)

        #return (ident,)
        #callback(ident)



    def __handle_open_display_with_id(self, callback, path, ident):

        self.__feedback.set_loading(path)

        def open_window():
            self.__feedback.show()

        def close_window(*args):
            self.__feedback.hide()

        def f(*args):
            return (ident, path, self.__DISPLAY_TYPE_WINDOW)

        def g(*args):
            return ident

        self.__pipeline([open_window, f, self.__add_display, close_window,
                         g, callback])
        #gobject.timeout_add(100, self.__add_display, callback, ident, path,
        #                    self.__DISPLAY_TYPE_WINDOW)
        #callback()



    def __handle_open_plug(self, callback, path):

        import time
        ident = str(time.time())
        plug = self.__add_display(ident, path, self.__DISPLAY_TYPE_PLUG)
        xid = plug.get_xembed_id()

        #return (ident, xid)
        callback(ident, xid)



    def __handle_close_display(self, callback, ident):

        self.__remove_display(ident)
        callback()


    def __handle_version(self, callback):

        #return (NAME, VERSION)
        callback(NAME, VERSION)



    def __handle_about(self, callback):

        callback("%s %s" % (NAME, VERSION),
                 "%s" % (COPYRIGHT.replace(u"\xa9", u"(C)"),),
                 "This software is licensed under the terms of the GNU GPL.")



    def __handle_shutdown(self, callback, *args):

        #self.__save_positions()

        for ident, display in self.__open_displays.items():
            try:
                display.remove_display()
            # what kind of exception are we expecting here?
            except:
               log("Could not remove desklet \"%s\"!" % (display,))
        # FIXME circumvent: 'gtk.ImageMenuItem' object is not callable
        try:
            callback()
        except:
            pass
        gtk.main_quit()



    def __handle_manage(self, *args):

        cmd = os.path.join(HOME, "gdesklets-shell")
        os.system(cmd + " &")


    def __handle_configger(self, *args):

        self.__configger.show()



    def __handle_about_dialog(self, *args):

        self.__about_dialog.show()



    def __handle_show_log(self, *args):

        cmd = os.path.join(HOME, "gdesklets-logview")
        os.system(cmd + " &")


    def __handle_displays(self, callback):

        #return self.__open_displays.keys()
        callback(self.__open_displays.keys())



    def __handle_display_list(self, callback):

        #return self.__open_displays.values()
        callback(*self.__open_displays.values())


    def __handle_geometry(self, callback, ident):

        display = self.__open_displays[ident]

        #return display.get_geometry()
        callback(*display.get_geometry())


    def __handle_set_remove_command(self, callback, command):

        # only allow setting the remove command once for security reasons
        if (not self.__remove_command):
            self.__remove_command = command

        callback()



    #
    # Executes the given commands in a pipeline. A command takes the return
    # values of the previous command as its arguments.
    #
    def __pipeline(self, line, *args):

        next = line.pop(0)
        try:
            retval = next(*args)
        except:
            from utils.ErrorFormatter import ErrorFormatter
            log(ErrorFormatter().format(sys.exc_info()))
            retval = ()

        if (line):
            if (not type(retval) in (type(()), type([]))):
                retval = (retval,)
            gobject.idle_add(self.__pipeline, line, *retval)




#
# Class for giving feedback while loading a display.
#
# TODO: move into its own file
#
class LoadingFeedback(gtk.Window):

    def __init__(self):

        gtk.Window.__init__(self, gtk.WINDOW_POPUP)
        self.set_position(gtk.WIN_POS_CENTER)

        frm = gtk.Frame()
        frm.set_shadow_type(gtk.SHADOW_OUT)
        frm.show()
        self.add(frm)
        
        box = gtk.HBox(spacing = 12)
        box.set_border_width(12)
        box.show()
        frm.add(box)

        self.__icon = gtk.Image()
        box.pack_start(self.__icon)

        self.__label = gtk.Label("")
        self.__label.show()
        box.pack_end(self.__label)



    def set_loading(self, path):

        from MetaData import MetaData
        try:
            meta = MetaData(path)
            preview = meta.get(meta.KEY_PREVIEW)
        except:
            preview = ""

        dirname = os.path.dirname(path)
        icon = os.path.join(dirname, preview)

        self.__label.set_markup("<big><b>Loading:</b>\n%s</big>" % path)

        self.__icon.hide()
        if (preview):
            try:
                # for him who loves small try blocks: you don't gain anything
                # from splitting up this block except for badly readable code ;)
                from utils import vfs
                data = vfs.read_entire_file(icon)
                loader = gtk.gdk.PixbufLoader()
                loader.write(data, len(data))
                loader.close()
                pbuf = loader.get_pixbuf()

                # scale icon down while preserving aspect ratio
                width = pbuf.get_width()
                height = pbuf.get_height()
                scale = 48 / float(height)
                width *= scale
                if (width > 96): width = 96
                
                if (abs(scale - 1.0) > 0.001):
                    pbuf = pbuf.scale_simple(int(width), 48, 3)

                self.__icon.set_from_pixbuf(pbuf)
                self.__icon.show()

            except:
                pass

        #end if

        self.resize(10, 10)
        self.set_size_request(-1, -1)



from DisplayTarget import DisplayTarget
from utils.DOM import DOM
from utils import svg
from utils.datatypes import *
from layout import Unit
from utils import dialog
from utils import vfs
import utils

import gtk



#
# Target where sensors can draw on.
#
class TargetCanvas(DisplayTarget):

    def __init__(self, name, parent):

        # a mini DOM for accessing the SVG data
        self.__dom = None

        # the previous size of the widget; used to detect resizings
        self.__old_size = (0, 0)

        # the size of the image
        self.__image_size = (100, 100)


        DisplayTarget.__init__(self, name, parent)

        self.__widget = gtk.Image()
        #self.__widget.connect("expose-event", self.__on_expose)
        self.__widget.show()

        # the "graphics" property is not readable because otherwise it could
        # be used to spy out files on the user's system after loading them into
        # "uri"
        self._register_property("graphics", TYPE_STRING,
                                self._setp_graphics, None)
        self._register_property("dom", TYPE_OBJECT,
                                None, self._getp_dom)
        self._register_property("uri", TYPE_STRING,
                                self._setp_uri, self._getp)

        self._setp("graphics", "")

        # watch the element for geometry changes
        self.add_observer(self.__on_observe_size)



    def get_widget(self): return self.__widget


    def delete(self):

        del self.__dom
        del self.__widget
        DisplayTarget.delete(self)



    #
    # Observer for size.
    #
    def __on_observe_size(self, src, cmd, *args):

        x, y, w, h = src.get_geometry()
        if (cmd == src.OBS_GEOMETRY and
            (w.as_px(), h.as_px()) != self.__old_size):
            utils.request_call(self.__redraw)
            self.__old_size = (w.as_px(), h.as_px())


        

    #
    # Transforms the given coordinates into buffer space.
    #
    def __transform_coords(self, x, y):

        width, height = self.get_geometry()[2:4]
        tx = (width.as_px() / 2.0)  * (1.0 + float(x))
        ty = (height.as_px() / 2.0) * (1.0 - float(y))

        return (tx, ty)


    def __make_style(self, foreground, fill):

        s = "stroke:" + foreground
        if (fill): s+= ";fill:" + foreground
        else: s+= ";fill:none"
        out = "style='%s'" % s
        return out


    #
    # Performs the given drawing operations. This is used for backwards
    # compatibility. New code should directly send SVG data.
    #
    def __draw_svg(self, commands):

        w, h = self.get_geometry()[2:4]
        out = "<svg width='%d' height='%d'>" % (w.as_px(), h.as_px())
        current_fg = "rgb(0, 0, 0)"
        for c in commands:
            if (not c.strip()): continue
            parts = c.split()

            cmd, args = parts[0], parts[1:]

            if (cmd == "color"):
                color = args[0]
                gdkcolor = gtk.gdk.color_parse(color)
                current_fg = "rgb(%d, %d, %d)" \
                 % (gdkcolor.red >> 8, gdkcolor.green >> 8, gdkcolor.blue >> 8)

            elif (cmd == "line"):
                x1, y1, x2, y2 = args
                x1, y1 = self.__transform_coords(x1, y1)
                x2, y2 = self.__transform_coords(x2, y2)
                style = self.__make_style(current_fg, False)
                out += "<line x1='%f' y1='%f' x2='%f' y2='%f' %s/>" \
                       % (x1, y1, x2, y2, style)

            elif (cmd == "polygon"):
                fill = int(args[-1])
                style = self.__make_style(current_fg, fill)
                points = [ self.__transform_coords(args[i], args[i+1])
                           for i in range(0,len(args)-1, 2) ]
                if (points): path = "M%f %f " % (points.pop(0))
                while (points):
                    path += "L%f %f " % (points.pop(0))
                out += "<path d='%s' %s/>" % (path, style)

            elif (cmd == "rectangle"):
                x1, y1, x2, y2, fill = args
                style = self.__make_style(current_fg, fill)
                x1, y1 = self.__transform_coords(x1, y1)
                x2, y2 = self.__transform_coords(x2, y2)
                w = x2 - x1
                h = y2 - y1
                out += "<rect x='%f' y='%f' width='%f' height='%f' %s/>" \
                       % (x, y, w, h, style)

            #end if
        #end for
        out += "</svg>"
        self.__render(out)


    #
    # Renders the given SVG data.
    #
    def __render(self, data):

        def f():
            utils.request_call(self.__redraw)

        if (not data): return


        self.__dom = DOM(data).get_root()
        if (not self.__dom): return

        self.__dom.set_update_handler(f)

        # check if the SVG has dynamic or static size
        if (self.__dom.has_key("width") and self.__dom.has_key("height")):
            try:
                width = int(float(self.__dom["width"]))
                height = int(float(self.__dom["height"]))
                self.__image_size = (width, height)
            except:
                # FIXME: convert SVG units to gDesklets units in case units
                # are used
                pass

        self.__redraw()
        

    #
    # Redraws the canvas.
    #
    def __redraw(self):

        if (not self.__dom): return

        w, h = self.__widget.size_request()
        imgw, imgh = self.__image_size
        
        # crappy SVG needs the size to be given; just set it here so that it
        # dynamically takes the correct size
        self.__dom["width"] = str(w or 100)
        self.__dom["height"] = str(h or 100)

        # so that's why the XML parser inserted an empty <g> node... :)
        g = self.__dom.get_children()[0]
        g["transform"] = "scale(%f, %f)" % (float(w) / imgw,
                                            float(h) / imgh)

        svg.render(self.__widget, w, h, str(self.__dom))


    #
    # "graphics" property.
    #
    def _setp_graphics(self, key, value):

        # native SVG
        if (value and value.lstrip()[0] == "<"):
            self.__render(value)

        # legacy graphics language
        else:
            value = value.split(",")
            self.__draw_svg(value)

        self._setp(key, value)


    #
    # Returns the DOM object of the graphics.
    #
    def _getp_dom(self, key): return self.__dom


    #
    # Loads SVG from the given URI.
    #
    def _setp_uri(self, key, value):

        uri = self._get_display().get_full_path(value)
        try:
            data = vfs.read_entire_file(uri)
        except:
            return

        self.__render(data)
        self._setp(key, value)



from shell.Plugin import Plugin

import gtk


#
# Plugin for the menu bar. The bar can easily be extended through other plugins
# by putting items into the menu tree.
#
class UI_Menu(Plugin):


    def init(self):
        self.__menu_tree = []
        self.__menu_bar = gtk.MenuBar()
        self.__menu_bar.show()

        self.__widget = self.__menu_bar

        shell = self._get_plugin("UI_Shell")
        shell.set_menu(self.__widget)



    #
    # Returns the parent node and the index of the node of the given path.
    #
    def __get_node(self, path):

        parts = path.split("/")
        parent = [("", self.__menu_bar, self.__menu_tree)]
        index = 0

        itempath = ""
        for p in parts:
            nil, nil, parent = parent[index]
            if (not p):
                index = len(parent)
            else:
                index = 0
                found = False
                for cname, nil, children in parent:
                    if (cname == p):
                        found = True
                        break
                    index += 1
                #end for
                if (not found):
                    self.insert(itempath, p)
                    return self.__get_node(path)
                
            itempath += p + "/"
            #end if
        #end for

        return (parent, index)


    #
    # Sets the given node to contain the given item.
    #
    def __set_node(self, path, item):

        tmp = path.split("/")
        tmp.pop()
        ppath = "/".join(tmp)
        if (not ppath):
            submenu = self.__menu_bar
        else:
            parent, index = self.__get_node(ppath)
            nil, pitem, nil = parent[index]
            submenu = pitem.get_submenu()

            if (not submenu):
                submenu = gtk.Menu()
                pitem.set_submenu(submenu)

        parent, index = self.__get_node(path)
        name, nil, children = parent[index]
        parent[index] = (name, item, children)
        submenu.insert(item, index)
        


    #
    # Inserts the given node into the tree as an empty node.
    #
    def insert(self, path, name):

        parent, index = self.__get_node(path)
        parent.insert(index + 1, (name, None, []))
        


    #
    # Returns the children of the given node.
    #
    def list(self, path):

        if (path[-1] != "/"): path += "/"
        parent, index = self.__get_node(path)
        ret = []
        for name, nil, nil in parent:
            ret.append(name)

        return ret


    #
    # Sets the given node to be an insertion slot.
    #
    def set_slot(self, path):

        item = gtk.MenuItem()
        return self.__set_node(path, item)



    #
    # Sets the given node to be a plain menu item.
    #
    def set_item(self, path, icon, label, callback, *args):

        if (icon):
            if (not label):
                item = gtk.ImageMenuItem(icon)
            else:
                img = gtk.Image()
                if (icon.startswith("gtk") or icon.startswith("gnome")):
                    img.set_from_stock(icon, gtk.ICON_SIZE_MENU)
                else:
                    img.set_from_file(icon)
                item = gtk.ImageMenuItem(label)
                item.set_image(img)
        else:
            item = gtk.MenuItem(label)

        item.show()
        if (callback):
            def f(*nil): callback(*args)
            item.connect("activate", f)

        self.__set_node(path, item)



    #
    # Sets the given node to be a separator.
    #
    def set_separator(self, path):

        item = gtk.MenuItem()
        item.show()
        return self.__set_node(path, item)



    #
    # Sets the given node to be a checkbox menu item.
    #
    def set_check_item(self, path, label, callback, *args):

        item = gtk.CheckMenuItem(label)
        item.show()
        if (callback):
            def f(*nil): callback(*args)
            item.connect("activate", f)

        return self.__set_node(path, item)



    #
    # Sets the checked state of a checkbox.
    #
    def set_checked(self, path, value):

        parent, index = self.__get_node(path)
        nil, item, nil = parent[index]
        item.set_active(value)



    #
    # Removes the given node from the menu tree.
    #
    def remove_item(self, path):

        tmp = path.split("/")
        tmp.pop()
        ppath = "/".join(tmp)
        if (not ppath):
            submenu = self.__menu_bar
        else:
            parent, index = self.__get_node(ppath)
            nil, pitem, nil = parent[index]
            submenu = pitem.get_submenu()

            if (not submenu):
                submenu = gtk.Menu()
                pitem.set_submenu(submenu)

        parent, index = self.__get_node(path)
        name, item, nil = parent[index]
        if (item):
            submenu.remove(item)
        del parent[index]
    

def get_class(): return UI_Menu



from ConfigWidget import ConfigWidget

import gtk

class ConfigTitle(ConfigWidget):

    def __init__(self, name, getter, setter, caller):

        ConfigWidget.__init__(self, name, getter, setter, caller)


    def get_widgets(self):

        self.__label = gtk.Label("")
        self.__label.show()
        align = gtk.Alignment(0.0, 1.0, 0.0, 0.0)
        align.show()
        align.add(self.__label)

        return (align,)


    def _set_label(self, value): self.__label.set_markup("<b>" + value + "</b>")



class HashPQueue:
    """
    Hashing Priority Queue
    (c) 2001, 2005 Martin Grimme  <martin@gdesklets.org>

    This class implements a hashing priority queue. It is similar to a heap
    data structure, but allows to remove arbitrary elements, not only the
    topmost one.
    """

    
    # The comparators for the heap.
    LESS_THAN = 0
    GREATER_THAN = 1



    def __init__(self, comparator = LESS_THAN):
        """
        Creates a new HashPQueue object with the given comparator.
        """

        self.__heap = []
        self.__id2pos = {}
        self.__pos2id = {}
        self.__index_counter = 0
        self.__comparator = comparator



    def __compare(self, v1, v2):
        """
        Compares the given two values by applying the comparator.
        Returns whether the comparator returned True.
        """

        if (self.__comparator == self.LESS_THAN):
            return (v1 < v2)
        if (self.__comparator == self.GREATER_THAN):
            return (v1 > v2)



    def __swap(self, i1, i2):
        """
        Swaps the two elements given by their index values.
        """

        obj1 = self.__heap[i1]
        obj1_id = self.__pos2id[i1]
        obj2 = self.__heap[i2]
        obj2_id = self.__pos2id[i2]

        self.__id2pos[obj1_id] = i2
        self.__pos2id[i2] = obj1_id

        self.__id2pos[obj2_id] = i1
        self.__pos2id[i1] = obj2_id
        
        self.__heap[i1], self.__heap[i2] = self.__heap[i2], self.__heap[i1]



    def insert(self, object):
        """
        Inserts a new object into the queue. The object must be comparable.
        Returns the identifier for the object by which it can be removed again.
        """

        ident = self.__index_counter
        self.__index_counter += 1

        position = len(self.__heap)
        self.__heap.append(object)
        self.__id2pos[ident] = position
        self.__pos2id[position] = ident
        self.__heapify_upwards(position)

        return ident



    def pick(self):
        """
        Returns and removes the first object from the queue, or None if the
        queue is empty.
        """

        obj = self.top()
        if (obj != None): self.remove(self.__pos2id[0])

        return obj



    def remove(self, ident):
        """
        Removes the object with the given identifier from the queue.
        """

        if (ident in self.__id2pos):
            index = self.__id2pos[ident]

            obj_index = len(self.__heap) - 1
            obj = self.__heap[obj_index]
            obj_id = self.__pos2id[obj_index]
            
            del self.__id2pos[ident]
            del self.__pos2id[index]

            self.__heap.pop()
            if (index != obj_index):
                del self.__pos2id[obj_index]
                self.__id2pos[obj_id] = index
                self.__pos2id[index] = obj_id
                self.__heap[index] = obj
                self.__heapify_downwards(index)



    def top(self):
        """
        Returns the first object in the queue without removing it, or None if
        the queue is empty.
        """
        
        # return 'None' if the queue is empty
        if (not self.__heap):
            return None
        else:
            return self.__heap[0]



    def has(self, object):
        """
        Returns whether the queue contains the given object.
        """

        return object in self.__heap
    


    def __heapify_upwards(self, position):
        """
        Restores the heap structure on the heap from bottom up, starting at the
        given position.
        """

        # terminate the recursion as soon as the top of the heap has been
        # reached
        if (position == 0):
            return

        # get the index of the parent node
        parent = (position - 1) / 2

        # get the elements from the nodes
        element1 = self.__heap[position]
        element2 = self.__heap[parent]

        # compare them and swap them if necessary
        if (self.__compare(element1, element2)):
            self.__swap(position, parent)

        # continue heapifying
        self.__heapify_upwards(parent)



    def __heapify_downwards(self, position):
        """
        Restores the heap structure on the heap from top down, starting at the
        given position.
        """

        length = len(self.__heap)

        # get the indices of the child nodes
        child1 = 2 * position + 1
        child2 = 2 * position + 2

        # if there are no children, terminate the recursion
        if (child1 >= length):
            return

        # if there is only one child, take this one
        if (child1 < length and child2 >= length):
            # get the elements from the nodes
            element1 = self.__heap[position]
            element2 = self.__heap[child1]
            child = child1
            
        # if there are two children, take the one complying to the comparator
        if (child1 < length and child2 < length):
            # get the elements from the nodes
            element1 = self.__heap[position]
            childelem1 = self.__heap[child1]
            childelem2 = self.__heap[child2]

            # find the appropriate child element
            if (self.__compare(childelem1, childelem2)):
                element2 = childelem1
                child = child1
            else:
                element2 = childelem2
                child = child2

            
        # compare the nodes and swap them if necessary
        if (not self.__compare(element1, element2)):
            self.__swap(position, child)

        # continue heapifying
        self.__heapify_downwards(child)





#
# Sort a list by using the heapsort algorithm to demonstrate the heap's usage.
#
if (__name__ == "__main__"):
    items = [0, 5, 6, 3, 56, 23, 89, 543, 12, 546,
             76, 34, 5, 3, 4, 2, 4, 6, 3, 30]

    light_heap = HashPQueue(HashPQueue.LESS_THAN)
    heavy_heap = HashPQueue(HashPQueue.GREATER_THAN)

    for i in items:
        light_heap.insert(i)
        heavy_heap.insert(i)

    ascend = []
    while (1):
        object = light_heap.pick()
        print object
        if (object == None):
            break
        else:
            ascend.append(object)

    descend = []
    while (1):
        object = heavy_heap.pick()
        if (object == None):
            break
        else:
            descend.append(object)

    print "\nThe items to sort:\n", items
    print "\nSorted in ascending order:\n", ascend
    print "\nSorted in descending order:\n", descend
    print



from TargetGroup import TargetGroup
from utils.datatypes import *
from layout import Unit


class TargetDisplay(TargetGroup):

    def __init__(self, name, parent):

        TargetGroup.__init__(self, name, parent)

        self._register_property("desktop-borders", TYPE_UNIT_LIST,
                                self._setp_container_stuff, self._getp)

        self._register_property("window-flags", TYPE_LIST,
                                self._setp_container_stuff, self._getp)
        self._register_property("shape", TYPE_STRING,
                                self._setp_container_stuff, self._getp)
        self._register_property("icon", TYPE_STRING,
                                self._setp_container_stuff, self._getp)
        self._register_property("title", TYPE_STRING,
                                self._setp_container_stuff, self._getp)

        self._register_property("path", TYPE_STRING, None, self._getp)
        self._setp("path", self._get_display().get_path())

        self.set_prop("x", Unit.Unit())
        self.set_prop("y", Unit.Unit())


        self._setp_container_stuff("window-flags", ["managed", "decorated"])



    #
    # Container properties.
    #
    def _setp_container_stuff(self, key, value):

        self._get_display().set_prop(key, value)
        self._setp(key, value)



from ConfigWidget import ConfigWidget
from utils.datatypes import *

import gtk

class ConfigFloat(ConfigWidget):

    def __init__(self, name, getter, setter, caller, int_only = False):

        self.__int_only = int_only

        ConfigWidget.__init__(self, name, getter, setter, caller)

        self._register_property("min", TYPE_FLOAT, self._setp_lhd,
                                self._getp, 0.0, doc = "Lower bound")
        self._register_property("max", TYPE_FLOAT, self._setp_lhd,
                                self._getp, 9999.0, doc = "Upper bound")
        self._register_property("digits", TYPE_INT, self._setp_lhd,
                                self._getp, 2,
                                doc = "Numbers of decimal digits")
        self._register_property("increment", TYPE_FLOAT, self._setp_increment,
                                self._getp, 1, doc = "Size of increments")
        self._register_property("value", TYPE_FLOAT, self._setp_value,
                                self._getp, 0.0, doc = "Value")



    def get_widgets(self):

        low = self.get_prop("min")
        high = self.get_prop("max")
        digits = self.get_prop("digits")

        self.__label = gtk.Label("")
        self.__label.show()
        entry = gtk.Entry()
        entry.show()
        align = gtk.Alignment(0.0, 0.0, 0.0, 0.0)
        align.show()
        align.add(self.__label)

        adjustment = gtk.Adjustment(0, low, high, 1, 1, 0)
        if (self.__int_only):
            self.__spin_button = gtk.SpinButton(adjustment, 1, 0)
        else:
            self.__spin_button = gtk.SpinButton(adjustment, 1, digits)
        self.__spin_button.set_numeric(True)
        self.__spin_button.show()

        self.__spin_button.connect("value-changed", self.__on_change)

        return (align, self.__spin_button)


    def __on_change(self, src):

        if (self.__int_only):
            value = src.get_value_as_int()
        else:
            value = src.get_value()
            
        self._set_config(value)


    def _set_enabled(self, value): self.__spin_button.set_sensitive(value)
    def _set_label(self, value): self.__label.set_text(value)


    def _setp_lhd(self, key, value):

        low, high = self.__spin_button.get_range()
        if (key == "min"):
            self.__spin_button.set_range(value, high)
        elif (key == "max"):
            self.__spin_button.set_range(low, value)
        elif (key == "digits"):
            self.__spin_button.set_digits(value)

        self._setp(key, value)


    def _setp_increment(self, key, value):

        self.__spin_button.set_increments(value, value * 10);
        self._setp(key, value)

    def _setp_value(self, key, value):

        self.__spin_button.set_value(value)
        self._setp(key, value)



from layout.LayoutObject import LayoutObject
from utils.datatypes import *
from utils.Struct import Struct
from layout import Unit
import utils
from DataTarget import DataTarget

from urllib import unquote

import gobject
import gtk


#
# Abstract class for DisplayTargets.
#
class DisplayTarget(DataTarget):

    __slots__ = ('__layout_object', '_geometry_lock', '__anchor',
                 '__had_action', '__action_stamp', '__pushed_cursor')

    # observer actions
    OBS_GEOMETRY = 0

    # user actions
    ACTION_CLICK = "click"
    ACTION_DOUBLECLICK = "doubleclick"
    ACTION_PRESS = "press"
    ACTION_RELEASE = "release"
    ACTION_MENU = "menu"
    ACTION_SCROLL = "scroll"
    ACTION_ENTER = "enter"
    ACTION_LEAVE = "leave"
    ACTION_MOTION = "motion"
    ACTION_FILE_DROP = "file-drop"
    ACTION_LINK_DROP = "link-drop"
    ACTION_KEY_PRESS = "key-press"
    ACTION_KEY_RELEASE = "key-release"

    # placement anchors
    ANCHOR_NW = "nw"
    ANCHOR_N = "n"
    ANCHOR_NE = "ne"
    ANCHOR_E = "e"
    ANCHOR_SE = "se"
    ANCHOR_S = "s"
    ANCHOR_SW = "sw"
    ANCHOR_W = "w"
    ANCHOR_CENTER = "center"


    # what we accept for drag & drop
    __DND_FILE = [("text/uri-list", 0, 0)]
    __DND_LINK = [("x-url/http", 0, 0), ("_NETSCAPE_URL", 0, 0)]


    def __init__(self, name, parent):

        # the action stamp helps us to detect ENTER and LEAVE events
        self.__action_stamp = 0
        
        # the layout object
        self.__layout_object = parent.new_layout_object()
        self.__layout_object.set_callback(self.__geometry_callback)
        self.__layout_object.set_action_callback(self.__action_callback)

        # the currently pushed mouse cursor
        self.__pushed_cursor = None
        
        # lock for geometry computations
        self._geometry_lock = True

        # the placement anchor
        self.__anchor = self.ANCHOR_NW

        # the value of the last notify_handle_action call
        self.__had_action = False


        DataTarget.__init__(self, name, parent)

        for prop, datatype in [("x", TYPE_UNIT),
                               ("y", TYPE_UNIT),
                               ("width", TYPE_UNIT),
                               ("height", TYPE_UNIT)]:
            self._register_property(prop, datatype,
                                    self._setp_geometry, self._getp_geometry)
        self.set_prop("x", Unit.ZERO)
        self.set_prop("y", Unit.ZERO)
        self.set_prop("width", Unit.Unit())
        self.set_prop("height", Unit.Unit())

        self._register_property("relative-to", TYPE_LIST,
                                self._setp_relative_to, self._getp)

        self._register_property("anchor", TYPE_STRING,
                                self._setp_anchor, self._getp)

        self._register_property("cursor", TYPE_STRING,
                                self._setp_cursor, self._getp)

        self._register_property("menu", TYPE_OBJECT,
                                self._setp_menu, None)

        self._register_property("visible", TYPE_BOOL,
                                self._setp_visible, self._getp)
        self._setp("visible", True)

        for action in (self.ACTION_ENTER,
                       self.ACTION_LEAVE,
                       self.ACTION_CLICK,
                       self.ACTION_PRESS,
                       self.ACTION_RELEASE,
                       self.ACTION_SCROLL,
                       self.ACTION_FILE_DROP,
                       self.ACTION_LINK_DROP,
                       self.ACTION_DOUBLECLICK,
                       self.ACTION_MOTION,
                       self.ACTION_MENU,
                       self.ACTION_KEY_PRESS,
                       self.ACTION_KEY_RELEASE):
            self._register_action(action)


    #
    # Creates and returns a new layout object as a child of this element's
    # layout object.
    #
    def new_layout_object(self):

        return self.__layout_object.new_child()


    #
    # Returns the layout object of this element.
    #
    def get_layout_object(self):

        return self.__layout_object


    #
    # Callback handler for changes in the geometry.
    #
    def __geometry_callback(self, src, x, y, w, h):

        self.get_widget().set_size_request(w.as_px(), h.as_px())
        utils.request_call(self.update_observer, self.OBS_GEOMETRY)



    #
    # Callback handler for user actions.
    #
    def __action_callback(self, src, x, y, stamp, action, event):

        self.detect_enter(stamp)

        call = self.get_action_call(action)
        if (call):
            utils.request_call(self.handle_action, action, x, y, event)



    #
    # Detects and emits ENTER events.
    #
    def detect_enter(self, stamp):

        if (not self.__action_stamp):
            # send enter event
            action = self.ACTION_ENTER
            if (self.get_action_call(action)):
                self.handle_action(action, Unit.ZERO, Unit.ZERO, Struct())

            # change cursor
            cursor = self._getp("cursor")
            if (cursor):
                self._get_display().push_cursor(cursor)
                self.__pushed_cursor = cursor
                
        self.__action_stamp = stamp


    #
    # Detects and emits LEAVE events.
    #
    def detect_leave(self, stamp):

        if (self.__action_stamp and stamp != self.__action_stamp):
            self.__action_stamp = 0
            # send leave event
            action = self.ACTION_LEAVE

            if (self.get_action_call(action)):
                self.handle_action(action, Unit.ZERO, Unit.ZERO, Struct())

            # revert cursor
            if (self.__pushed_cursor):
                self._get_display().pop_cursor(self.__pushed_cursor)
                self.__pushed_cursor = None



    def _is_active(self): return (self.__action_stamp != 0)



    def __on_file_drop(self, widget, context, x, y, data, info, time):
        ''' catch DND events and process them to send them to them
        main display, which forwards them to the sensor '''

        # get the display
        display = self._get_display()

        # tell the main display to send files and coordinates to the sensor
        files = [unquote(uri) for uri in data.data.split("\r\n") if uri != '']
        display.send_action(self, self.ACTION_FILE_DROP,
                            Struct(files = files, _args = [files]))


    def __on_link_drop(self, widget, context, x, y, data, info, time):
        ''' catch DND events and process them to send them to them
        main display, which forwards them to the sensor '''

        # get the display
        display = self._get_display()

        # tell the main display to send link and coordinates to the sensor
        links = [unquote(data.data.split("\n")[0])]
        display.send_action(self, self.ACTION_LINK_DROP,
                            Struct(links = links, _args = [links]))


    #
    # Returns whether this target is standalone, i.e. needs no parent.
    #
    def is_standalone(self): return False


    #
    # Returns the widget of this target.
    #
    def get_widget(self): raise NotImplementedError


    #
    # Returns the true coordinates of this target when the given coordinates
    # are the hotspot.
    #
    def get_anchored_coords(self, x, y, w, h):
        assert (isinstance(x, Unit.Unit))
        assert (isinstance(y, Unit.Unit))
        assert (isinstance(w, Unit.Unit))
        assert (isinstance(h, Unit.Unit))

        if (x.is_unset() or y.is_unset()):
            return (x, y)

        anchor = self.__anchor
        if (anchor in (self.ANCHOR_NW, self.ANCHOR_W, self.ANCHOR_SW)):
            ax = x
        elif (anchor in (self.ANCHOR_N, self.ANCHOR_CENTER, self.ANCHOR_S)):
            ax = x - (w / 2)
        else:
            ax = x - w

        if (anchor in (self.ANCHOR_NW, self.ANCHOR_N, self.ANCHOR_NE)):
            ay = y
        elif (anchor in (self.ANCHOR_W, self.ANCHOR_CENTER, self.ANCHOR_E)):
            ay = y - (h / 2)
        else:
            ay = y - h

        return (ax, ay)




    #
    # Returns the geometry (coordinates and size) of this target.
    #
    def get_geometry(self):

        x, y, w, h = self.__layout_object.get_real_geometry()
        if (self.get_prop("visible")):
            return (x, y, w, h)
        else:
            return (x, y, Unit.ZERO, Unit.ZERO)




    #
    # Returns the geometry from the user's point of view.
    #
    def get_user_geometry(self):

        return self.__layout_object.get_geometry() #self.__user_geometry




    #
    # Sets the position of this target.
    #
    def set_position(self, x, y):
        assert (isinstance(x, Unit.Unit))
        assert (isinstance(y, Unit.Unit))

        ox, oy, w, h = self.__layout_object.get_geometry()
        if ((x, y) != (ox, oy)):
            self.__layout_object.set_geometry(x = x, y = y)



    #
    # Sets the size of this target. Use this instead of set_size_request() in
    # targets to set the size manually.
    #
    def set_size(self, width, height):
        assert (isinstance(width, Unit.Unit))
        assert (isinstance(height, Unit.Unit))

        x, y, w, h = self.__layout_object.get_geometry()
        if ((w, h) != (width, height)):
            if (w.get_unit() != Unit.UNIT_PERCENT):
                self.__layout_object.set_geometry(width = width)
            if (h.get_unit() != Unit.UNIT_PERCENT):
                self.__layout_object.set_geometry(height = height)
            #self.__layout_object.set_geometry(width = width, height = height)



    def handle_action(self, action, px, py, event):
        assert (isinstance(px, Unit.Unit))
        assert (isinstance(py, Unit.Unit))

        # we need the pointer position relative to the widget, so we have to
        # setup a new event structure for some actions
        if (action in (self.ACTION_CLICK, self.ACTION_DOUBLECLICK,
                       self.ACTION_MOTION, self.ACTION_PRESS,
                       self.ACTION_RELEASE)):
            x, y = self.get_widget().get_pointer()
            nil, nil, w, h = self.get_geometry()
            ux = Unit.Unit(x, Unit.UNIT_PX)
            uy = Unit.Unit(y, Unit.UNIT_PX)
            if (w.as_px() > 0): ux.set_100_percent(w.as_px())
            if (h.as_px() > 0): uy.set_100_percent(h.as_px())
            event["x"] = ux
            event["y"] = uy
            # FIXME: remove eventually :)
            if (action == self.ACTION_MOTION): event["_args"] = [x, y]

        DataTarget.handle_action(self, action, px, py, event)



    #
    # Geometry properties.
    #
    def _setp_geometry(self, key, value):
        assert (isinstance(value, Unit.Unit))

        if (key == "x"):
            self.__layout_object.set_geometry(x = value)
            self._setp(key, value)

        elif (key == "y"):
            self.__layout_object.set_geometry(y = value)
            self._setp(key, value)

        elif (key == "width"):
            self.__layout_object.set_geometry(width = value)
            self._setp(key, value)

        elif (key == "height"):
            self.__layout_object.set_geometry(height = value)
            self._setp(key, value)


    def _setp_relative_to(self, key, value):

        name, mode = value
        if (mode == "x"): rx, ry = True, False
        elif (mode == "y"): rx, ry = False, True
        elif (mode == "xy"): rx, ry = True, True
        else: rx, ry = False, False

        def f():
            try:
                parent = self._get_parent()
                my_id = self._getp("id")
                if (not parent or not parent.get_child_by_id(my_id)):
                    return

                obj = self._get_parent().get_child_by_id(name)
                if (not obj):
                    raise UserError(_("Element \"%s\" does not exist") % name,
                                   _("The <tt>relative-to</tt> property "
                                     "requires a reference to an existing "
                                     "display element within the same parent "
                                     "container."))
                
                relative = obj.get_layout_object()
                
                self.__layout_object.set_relative_to(relative, rx, ry)
            except:
                import traceback; traceback.print_exc()
                pass

        # we have to delay because the relative might not be available at that
        # time
        utils.request_call(f)
        
        self._setp(key, value)


    def _setp_anchor(self, key, value):

        self.__anchor = value
        if (value in (self.ANCHOR_NW, self.ANCHOR_W, self.ANCHOR_SW)):
            ax = 0.0
        elif (value in (self.ANCHOR_N, self.ANCHOR_CENTER, self.ANCHOR_S)):
            ax = 0.5
        else:
            ax = 1.0

        if (value in (self.ANCHOR_NW, self.ANCHOR_N, self.ANCHOR_NE)):
            ay = 0.0
        elif (value in (self.ANCHOR_W, self.ANCHOR_CENTER, self.ANCHOR_E)):
            ay = 0.5
        else:
            ay = 1.0

        self.__layout_object.set_anchor(ax, ay)
        self._setp(key, value)


    def _getp_geometry(self, key):

        #if (self._is_geometry_locked()):
        #    unit = self._getp(key)
        #    return unit

        x, y, w, h = self.__layout_object.get_real_geometry()
        #dx, dy = self.__layout_object.get_anchor() #self.get_anchor()
        #x += dx
        #y += dy
        
        #parent_width, parent_height = self._get_parent_size()

        if (key == "x"):
            unit = x
            #unit.set_100_percent(parent_width.as_px())
        elif (key == "y"):
            unit = y
            #unit.set_100_percent(parent_height.as_px())
        elif (key == "width"):
            unit = w
            #unit.set_100_percent(parent_width.as_px())
        elif (key == "height"):
            unit = h
            #unit.set_100_percent(parent_height.as_px())

        return unit


    #
    # "visible" property.
    #
    def _setp_visible(self, key, value):

        if (value): self.get_widget().show()
        else: self.get_widget().hide()
        self.__layout_object.set_enabled(value)
        self._setp(key, value)


    #
    # "menu" property.
    #
    def _setp_menu(self, key, value):

        dsp = self._get_display()
        utils.request_call(dsp.open_menu, value)
        


    #
    # Action handlers.
    #
    def _setp__action(self, key, value):

        DataTarget._setp__action(self, key, value)
        if (key == "on-file-drop"):
            self.get_widget().drag_dest_set(gtk.DEST_DEFAULT_ALL,
                                            self.__DND_FILE,
                                            gtk.gdk.ACTION_COPY)
            self.get_widget().connect("drag_data_received",
                                      self.__on_file_drop)

        elif (key == "on-link-drop"):
            self.get_widget().drag_dest_set(gtk.DEST_DEFAULT_ALL,
                                            self.__DND_LINK,
                                            gtk.gdk.ACTION_COPY)
            self.get_widget().connect("drag_data_received",
                                      self.__on_link_drop)


    #
    # "cursor" property.
    #
    def _setp_cursor(self, key, value):

        self._setp(key, value)
        if (self.__pushed_cursor):
            self._get_display().pop_cursor(self.__pushed_cursor)
            self._get_display().pop_cursor(value)
            self.__pushed_cursor = value


    #
    # Unlocks the initial lock for geometry computations.
    # By locking the geometry engine initially, we can build up the display
    # without redundant geometry computations. The necessary computations are
    # done once when unlocking.
    #
    def unlock_geometry(self):

        self._geometry_lock = False
        #utils.request_call(self.adjust_geometry)
        #self.adjust_geometry()


    #
    # Returns whether the geometry engine is locked.
    #
    def _is_geometry_locked(self): return self._geometry_lock



import os
import gtop
import NullObject

import CPU
import Net
import Swap


class Arch:

    def __init__(self):

        self.__cpu_total = CPU.Total()

        self.__cpu_cpus  = [ CPU.CPU(i) for i in range(len(gtop.cpu().cpus)) ]

        self.__net_devices = {}

        self.__swap = Swap.Swap()

        self.__os, self.__name, self.__kernel = os.uname()[:3]


    def _init(self):

        self.__net_devices = dict( [ (iface, Net.Net(iface))
                                     for iface in self.net_devices() ] )

        self._init = Null



    def kernel_version(self):
        """
        @return : kernel which is currently running
        @rtype  : str
        """

        return self.__kernel


    def hostname(self):
        """
        @return : the machine's name
        @rtype  : str
        """

        return self.__name


    def operating_system(self):
        """
        @return : operating system which is currently running
        @rtype  : str
        """

        return self.__os


    def net_speed(self, dev):
        """
        @param dev: interface
        @type  dev: str

        @return : (speed in, speed out)
        @rtype  : tuple
        """

        try:
            value = self.__net_devices[dev].poll()
        except KeyError:
            log("Info:\n"
                "%s doesn't exist! Falling back to \"lo\"." % (dev,))
            value = self.__net_devices["lo"].poll()
        return value


    def net_state(self, dev):
        """
        @param dev: interface
        @type  dev: str

        @return : whether a network device is up or not (i.e. down)
        @rtype  : bool
        """

        return gtop.netload(dev).if_flags & gtop.NETLOAD_IF_FLAGS_UP


    def users(self):
        """
        @return : number of users.
        @rtype  : int
        """

        return len(os.popen("users").read().split())


    def cpu_load(self):
        """
        @return : cpu load
        @rtype  : float
        """

        return self.__cpu_total.poll()


    def cpu_count(self):
        """
        @return : number of installed CPUs
        @rtype  : int
        """

        return len(self.__cpu_cpus)


    def all_cpu_load(self):
        """
        @return : total load of installed CPUs
        @rtype  : list
        """

        return [ cpu.poll() for cpu in self.__cpu_cpus ]


    def swap_speed(self):
        """
        @return : (speed in, speed out)
        @rtype  : tuple
        """

        return self.__swap.poll()


    # The following functions will be overriden in a subclass (Generic, ...).
    def net_devices(self):
        return []


    def cpu_model(self):
        return ""


    def cpu_speed(self):
        return 0.0


    def cpu_cache(self):
        return 0


    def cpu_bogomips(self):
        return 0.0




from tiling import Tiling

import gtk



tiling = Tiling()


#
# Restricts the given coordinates to visible values.
#
def _crop_coords(x, y, width, height):

    scrwidth = gtk.gdk.screen_width()
    scrheight = gtk.gdk.screen_height()

    x = min(x, scrwidth - 1)
    y = min(y, scrheight - 1)

    #width = min(scrwidth - x, width)
    #height = min(scrheight - y, height)

    return (x, y, width, height)



#
# Captures the wallpaper image by accessing the background pixmap.
#
def get_wallpaper(widget, x, y, width, height):

    if (HAVE_WIN32): return

    x, y, width, height = _crop_coords(x, y, width, height)
    # get wallpaper
    try:
        pmap_id = get_wallpaper_id()
        widget.set_from_background(pmap_id, x, y, width, height)

    except NotImplementedError:
        widget.set_from_background(0, x, y, width, height)

    widget.render(width, height, 1, 1)



#
# Returns the ID of the background pixmap.
#
def get_wallpaper_id():

    try:
        root = gtk.gdk.get_default_root_window()
        ident = root.property_get("_XROOTPMAP_ID", "PIXMAP")[2][0]
        return long(ident)

    except:
        raise NotImplementedError



from ContainerTarget import ContainerTarget
from Layouter import Layouter, LayoutError
from utils.datatypes import *
from layout import Unit
import utils

import gtk


#
# Class for arrays of TargetDisplays.
#
class TargetArray(ContainerTarget):

    def __init__(self, name, parent):

        # the current layouter function
        self.__layouter = []

        # the class of the array elements
        self.__elementtype = None

        # the default settings of the array elements
        self.__elementsettings = None

        # children data of the array elements
        self.__elementchildren = []

        # the current elements of this array
        self.__children = []

        # mapping between unique IDs and targets
        self.__targets = {}

        ContainerTarget.__init__(self, name, parent)
        self.__layout = gtk.Fixed()
        self.__layout.show()

        self._register_property("length", TYPE_INT,
                                self._setp_length, self._getp_length)
        self._register_property("layout", TYPE_LIST,
                                self._setp_layout, self._getp)



    #
    # Returns the widget
    #
    def get_widget(self): return self.__layout



    #
    # Returns next child index
    #
    def get_next_child_index(self): return len(self.__children)



    def new_child(self, childtype, settings, children):

        self.__elementtype = childtype
        self.__elementsettings = settings
        self.__elementchildren = children

        # the array should be empty at the beginning but we need to create a
        # child so the array can connect to sensor output
        # remove eventually
        self.__add_child()
        self.__remove_child()



    def __on_observe_children(self, src, cmd):

        def f():
            x, y, w, h = src.get_geometry()
            self.__layout.move(src.get_widget(), x.as_px(), y.as_px())


        if (cmd == src.OBS_GEOMETRY and self.__children):
            utils.request_call(f)



    #
    # Adds an element to this array.
    #
    def __add_child(self):

        # use a unique ID
        settings = self.__elementsettings.copy()
        settings["id"] += "#" + str(len(self.__children))
        child = ContainerTarget.new_child(self, self.__elementtype,
                                          settings,
                                          self.__elementchildren)
        self.__children.append(child)

        x, y, w, h = child.get_geometry()
        self.__layout.put(child.get_widget(), x.as_px(), y.as_px())
        child.add_observer(self.__on_observe_children)

        if (not self._is_geometry_locked()): child.unlock_geometry()



    #
    # Removes an element from this array.
    #
    def __remove_child(self):
        assert(self.__children)

        child = self.__children.pop()
        self.__layout.remove(child.get_widget())
        child.remove_observer(self.__on_observe_children)

        self.get_layout_object().remove_child(child.get_layout_object())

        self._unregister_child(child)
        child.delete()



    #
    # Positions all elements of this array.
    #
    def __place_children(self):

        if (not self.__layouter): return
        parts = self.__layouter
        layout = parts[0]
        args = parts[1:]

        try:
            layouter = Layouter(layout)
        except LayoutError:
            return

        cnt = 0
        previous_id = ""
        for child in self.__children:
            x, y, w, h = child.get_geometry()
            rel, cx, cy = layouter.layout(cnt, args)
            cx = Unit.Unit(cx, Unit.UNIT_PX)
            cy = Unit.Unit(cy, Unit.UNIT_PX)

            cx, cy = child.get_anchored_coords(cx, cy, w, h)

            if (not rel):
                child.set_position(cx, cy)

            if (previous_id and rel):
                child.set_prop("relative-to", (previous_id, rel))

            previous_id = self.get_id_by_child(child)
            cnt += 1

        #utils.request_call(self.adjust_geometry)
        #self.adjust_geometry()



    # FIXME: remove eventually
    def distribute_sensor_output(self, sensor, indexes, key, value):

        index = int(indexes.pop(0))

        # add children, if necessary
        if (index >= len(self.__children)):
            while (index >= len(self.__children)): self.__add_child()
            self.__place_children()

        entries = self.__targets.get((sensor, index, key), [])
        for target, prop in entries:
            if (indexes):
                target.distribute_sensor_output(sensor, indexes[:],
                                                key, value)
            else:
                target.set_xml_prop(prop, value)



    #
    # "length" property.
    #
    def _setp_length(self, key, value):

        length = int(value)
        if (len(self.__children) != length):
            while (length > len(self.__children)): self.__add_child()
            while (length < len(self.__children)): self.__remove_child()
        self.__place_children()



    def _getp_length(self, key):

        return len(self.__children)



    #
    # "layout" property.
    #
    def _setp_layout(self, key, value):

        self.__layouter = value
        self.__place_children()
        self._setp(key, value)




import gtk


class HIGDialog(gtk.Dialog):

    def __init__(self, buttons, type = None, primary = "", secondary = "",
                 self_destroy = True):

        gtk.Dialog.__init__(self, "", None, gtk.WINDOW_TOPLEVEL, buttons)

        # HIG says don't show a separator
        self.set_property("has-separator", False)
        # HIG says don't show up in the taskbar
        self.set_property("skip-taskbar-hint", True)
        # HIG says dialog mustn't be resizable
        self.set_property("resizable", False)
        # HIG says border-width must be set to 6
        self.set_property("border-width", 6)

        self.vbox.set_property("spacing", 12)

        hbox = gtk.HBox()
        hbox.set_property("spacing", 12)
        hbox.set_property("border-width", 6)
        hbox.show()

        if (type):
            icon = gtk.Image()
            icon.set_property("yalign", 0.0)
            icon.set_from_stock(type, gtk.ICON_SIZE_DIALOG)
            icon.show()
            hbox.pack_start(icon, False, False, 0)

            self.vbox.pack_start(hbox, False, False, 0)

        if (primary != "" and secondary != ""):
            # according to the HIG, a dialog message has to look like this
            m = "<span weight=\"bold\" size=\"larger\">%(primary)s</span>\n" \
                "\n" \
                "%(secondary)s" % vars()

            label = gtk.Label()
            label.set_property("use-markup", True)
            label.set_property("selectable", True)
            label.set_property("wrap", True)
            label.set_property("yalign", 0.0)
            label.set_label(m)
            label.show()
            hbox.pack_start(label, False, False, 0)

            # check if the hbox is already added to the vbox
            if (not isinstance(self.vbox.get_children()[0], gtk.HBox)):
               self.vbox.pack_start(hbox, False, False, 0)

        if (self_destroy):
            def f(*args): self.destroy()
            self.connect_after("response", f)




#
# Client module for the gDesklets daemon. Use this module to talk to the
# daemon from external programs.
#

from main import DISPLAY, HOME, SOCKET_PATH
from main.remotecommands import *
import utils.xdr as xdr

import os
import sys
import socket
import time

from utils import i18n
import __builtin__
__builtin__._ = i18n.Translator("gdesklets")


class RemoteConnection:


    def __init__(self, path):

        if (HAVE_WIN32):
            self.__sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            self.__sock.connect(("localhost", 1979))
        else:
            self.__sock = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
            self.__sock.connect(path)



    def __send(self, command, *args):

        """
        Sends the given command to the given socket and returns the answer.
        """

        xdr.send(self.__sock, command, *args)

        return xdr.recv(self.__sock)



    def open_display(self, path):

        """
        Opens a display file.

        @param path: The (VFS-)path of the display file to open.
        @type  path: str

        @return: The ID of the new display.
        @rtype : str
        """

        return self.__send(COMMAND_OPEN_DISPLAY, path)[0]



    def open_display_with_id(self, path, ident):

        """
        Opens a display file with the specified ID.

        @param path: The (VFS-)path of the display file to open.
        @type  path: str
        @param ident: The ID of the display.
        @type  ident: str
        """

        return self.__send(COMMAND_OPEN_DISPLAY_WITH_ID, path, ident)



    def open_plug(self, path):

        """
        Opens a display file as an embeddable plug.

        @param path: The (VFS-)path of the display file to open.
        @type  path: str

        @return: (id, xid) The ID of the new display and the XID of the plug
                 which is cimpliant with the XEmbed protocol.
        @rtype : str
        """

        ident, xid = self.__send(COMMAND_OPEN_PLUG, path)

        return (ident, long(xid))



    def close_display(self, ident):

        """
        Closes a display with the specified ID.

        @param ident: The ID of the display.
        @type  ident: str
        """

        self.__send(COMMAND_CLOSE_DISPLAY, ident)


    def configure(self):

        """
        Opens the configuration dialog.
        """

        self.__send(COMMAND_CONFIGURE)


    def version(self):

        """
        Shows the gDesklets version.

        @return: Tuple with the name and version of the gDesklets daemon
        @rtype : tuple(str, str)
        """

        return self.__send(COMMAND_VERSION)



    def about(self):

        """
        Shows some additional information about gDesklets.

        @return: gDesklets information
        @rtype : tuple(str, str, str)
        """

        return self.__send(COMMAND_ABOUT)



    def shutdown(self):

        """
        Shuts down the daemon.

        """

        self.__send(COMMAND_SHUTDOWN)



    def displays(self):

        """
        Shows the IDs of running displays.

        @return: The list of the IDs of the currently open displays.
        @rtype : tuple
        """

        return self.__send(COMMAND_DISPLAYS)



    def display_list(self):

        """
        Shows a list of running displays.

        @return: The list of currently open displays.
        @rtype: list
        """

        return self.__send(COMMAND_DISPLAY_LIST)



    def geometry(self, ident):

        """
        Shows the coordinates and the width, height of the display specified by
        the given ID.

        @param ident: The ID of the display.
        @type  ident: str

        @return: Tuple of the x, y, width, height.
        @rtype : tuple(int, int, int, int)
        """

        x, y, w, h = self.__send(COMMAND_GEOMETRY, ident)

        return (int(x), int(y), int(w), int(h))



    def set_remove_command(self, cmd):

        """
        Sets the remove command which will be run whenever the user removes
        a display. This can only be set once in a daemon's lifetime for
        security reasons.

        @param cmd: The command.
        @type  cmd: str
        """

        self.__send(COMMAND_SET_REMOVE_COMMAND, cmd)



def daemon_is_running(display = None):

    """
    Tells if the daemon is running.

    @param display: display on which should be queried for the daemon
    @type  display: str

    @return: True if the daemon is running, False if not.
    @rtype : bool
    """

    if (not display):
        display = DISPLAY

    path = os.path.join(SOCKET_PATH, display)

    if (HAVE_WIN32 or os.path.exists(path)):
        try:
            connection = RemoteConnection(path)
            return (connection != None)
        except:
            pass

    return False



def get_daemon(display = None):

    """ Returns the daemon. If there is no daemon running, start a new one.

    @param display: display on which should be queried for the daemon
    @type  display: str

    @return:
    @rtype :
    """

    if (not display):
        display = DISPLAY

    daemon = None
    path = os.path.join(SOCKET_PATH, display)
    boottime = time.time()

    if (daemon_is_running(display)):
        try:
            connection = RemoteConnection(path)
            daemon = connection
        except:
            os.unlink(path)

    else:
        daemon_path = os.path.join(HOME, "gdesklets-daemon")

        args = " ".join([ a for a in sys.argv if a.startswith("--") ])

        if (HAVE_WIN32):
            os.system("start python \"%s\" %s" % (daemon_path, args))
        else:
            os.system("%s %s &" % (daemon_path, args))

        pos = 0
        delta = 1
        for i in range(500):
            left = pos
            right = 10 - pos
            mybar = " " * left + "###" + " " * right
            sys.stdout.write("\r" + _("Connecting to daemon [%s]") % mybar)
            sys.stdout.flush()
            pos += delta
            if (pos == 0 or pos == 10):
                delta = -delta
            try:
                connection = RemoteConnection(path)
                daemon = connection
                if (daemon):
                    break
            except socket.error:
                pass
            time.sleep(0.1)

        print "\r" + " " * 40 + "\r",

        if (not daemon):
            sys.exit(_("Cannot establish connection to daemon: timeout!\n"
                        "The log file might help you solving the problem."))

    elapsed = int(1000000*(time.time() - boottime))

    if (elapsed < 1000):
        print _("Connected to daemon in %d microseconds.") % (elapsed,)
    else:
        elapsed /= 1000
        print _("Connected to daemon in %d milliseconds.") % (elapsed,)

    return daemon




#
# Module for storing confidential data like passwords.
#
# This was modeled after the mozilla password storage system with some security
# improvements; in fact it looks totally different.
#
# This one uses a prefix where to store the key file. Don't ever let somebody
# read your key file!
# The keys are worthless without the key file and vice versa.
#

import utils


import random
import os
import base64



def _init_store(prefix):

    l = os.listdir(prefix)
    l.sort()
    if (not (l and l[0].startswith(".!pwstore"))):
        _create_store(prefix)



def _create_store(prefix):

    r1 = random.randrange(0xffffffL, 0x7fffffffL)
    r2 = random.randrange(0xffffffL, 0x7fffffffL)
    fname1 = ".!pwstore" + str(r1)
    fname2 = ".!" + str(r2)
    utils.makedirs(os.path.join(prefix, fname1))
    path = os.path.join(prefix, fname1, fname2)

    # is this equivalent to the commented code below? i think so
    # chars = [chr(a) for a in range(256)*16 if a!=26]
    # better :D
    chars = [chr(a) for a in range(256) if a!=26] * 16
    #chars = []
    #for i in xrange(4096):
    #    a = i % 256
    #    if (a == 26): continue
    #    chars.append(chr(a))
    #end for

    data = ""
    while chars:
        index = random.randrange(len(chars))
        c = chars.pop(index)
        data += c
    #end while

    fd = open(path, "w")
    fd.write(data)
    fd.close()
    os.chmod(path, 0400)



def _open_store(prefix):

    try:
        l = os.listdir(prefix)
        l.sort()
        fname1 = l[0]
        l = os.listdir(os.path.join(prefix, fname1))
        fname2 = l[0]

        path = os.path.join(prefix, fname1, fname2)
        data = open(path).read()

        return data

    except OSError:
        return



#
# Stores the given value and returns a key for retrieval.
#
def store(prefix, value):

    _init_store(prefix)
    data = _open_store(prefix)

    v64 = base64.encodestring(repr(value))
    key = u""
    for c in v64:
        index = data.index(c)
        key += unichr(index)
    #end for

    return repr(key)



#
# Returns the value for the given key.
#
def retrieve(prefix, key):

    if (not key): return ""

    _init_store(prefix)
    data = _open_store(prefix)

    try:
        v64 = ""
        for c in eval(key):
            v64 += data[ord(c)]
	#end for

	value = eval(base64.decodestring(v64))
	return value

    except base64.binascii.Error:
        return ""



from libdesklets.controls import Control
from libdesklets import hddtemp

from IHDDTemp import IHDDTemp


class HDDTemp(Control, IHDDTemp):

    def __init__(self):
        super(Control, self).__init__()
        self.__device = None


    def __get_device(self):
        assert self.__device
        return self.__device


    def __set_device(self, dev):
        assert isinstance(dev, str)
        assert dev in self.__get_available_devices()
        self.__device = dev


    def __get_poll(self):
        assert self.__device
        return hddtemp.poll(self.__device)

    
    def __get_available_devices(self):
        return hddtemp.available_devices()


    def __get_poll_all(self):
        return hddtemp.poll_all()



    available_devices   = property(fget = __get_available_devices)
    poll_all            = property(fget = __get_poll_all)
    device              = property(fget = __get_device,
                                   fset = __set_device)
    poll                = property(fget = __get_poll)




from utils.Observable import Observable

from tips import TIPS

import gtk


#
# Class for tip-of-the-day dialogs.
#
class TipOfTheDay(gtk.MessageDialog, Observable):

    OBS_TIP = 0
    OBS_TOGGLED = 1


    def __init__(self, index):

        self.__index = index - 1

        gtk.MessageDialog.__init__(self, None, gtk.DIALOG_MODAL,
                                   gtk.MESSAGE_INFO, gtk.BUTTONS_NONE,
                                   "gDesklets")
        self.set_title(_("Tip of the Day"))

        self.__label = self.vbox.get_children()[0].get_children()[1]
        # HACK HACK HACK -- make it work with gtk+-2.6.x
        if (not isinstance(self.__label, gtk.Label)):
            self.__label.remove(self.__label.get_children()[1])
            self.__label = self.__label.get_children()[0]

        self.__checkbox = gtk.CheckButton(_("Show tips at startup"))
        self.__checkbox.set_active(True)
        self.__checkbox.connect("toggled", self.__on_toggle)
        self.__checkbox.show()
        self.vbox.pack_end(self.__checkbox, 0, 0, 0)

        btn_prev = gtk.Button(stock = gtk.STOCK_GO_BACK)
        btn_prev.connect("clicked", self.prev_tip)
        btn_prev.show()

        btn_next = gtk.Button(stock = gtk.STOCK_GO_FORWARD)
        btn_next.connect("clicked", self.next_tip)
        btn_next.show()

        btn_close = gtk.Button(stock = gtk.STOCK_CLOSE)
        btn_close.connect("clicked", self.__on_close)
        btn_close.show()

        self.action_area.add(btn_prev)
        self.action_area.add(btn_next)
        self.action_area.add(btn_close)

        # make the close button the default button
        btn_close.set_flags(gtk.CAN_DEFAULT)
        self.set_default(btn_close)
        self.set_focus(btn_close)


    def set_show(self, value):

        self.__checkbox.set_active(value)


    def __on_close(self, src):

        self.hide()


    def __on_toggle(self, src):

        value = src.get_active()
        self.update_observer(self.OBS_TOGGLED, value)


    def next_tip(self, *args):

        self.__index += 1
        if (self.__index >= len(TIPS)): self.__index = 0

        self.__label.set_markup(_(TIPS[self.__index]).strip())
        self.update_observer(self.OBS_TIP, self.__index + 1)


    def prev_tip(self, *args):

        self.__index -= 1
        if (self.__index < 0): self.__index = len(TIPS) + self.__index

        self.__label.set_markup(_(TIPS[self.__index]).strip())
